# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-08-22 22:43:15
# @Last modified by:   Brian Cherinka
# @Last modified time: 2018-11-12 17:06:25

from __future__ import absolute_import, division, print_function

import copy as copy_mod
import inspect
import os
import warnings

import numpy as np
import six
import yaml
import yamlordereddictloader
from astropy.table import Table
from fuzzywuzzy import fuzz, process
from sqlalchemy_utils import get_hybrid_properties

from marvin import config
from marvin.core.exceptions import MarvinError, MarvinUserWarning
from marvin.utils.datamodel import DataModelList
from marvin.utils.datamodel.maskbit import get_maskbits
from marvin.utils.general.structs import FuzzyList


if config.db:
    from marvin.utils.datamodel.query.forms import MarvinForm
else:
    MarvinForm = None


__ALL__ = ('QueryDataModelList', 'QueryDataModel')


query_params = None

PARAM_CACHE = {}


def get_allowed_releases():
    ''' get the supported API / web MaNGA releases '''
    min_release = getattr(config, '_min_web_release', None)
    public = config.access == 'public' or os.environ.get(
        'PUBLIC_SERVER', None) is True
    allowed_releases = config.get_allowed_releases(
        public=public, min_release=min_release)
    return allowed_releases


def is_supported_release(release):
    ''' check a release against the list of supported API releases '''
    allowed_releases = get_allowed_releases()
    return release in allowed_releases.keys()


class QueryDataModel(object):
    """ A class representing a Query datamodel """


    def __init__(self, release, groups=[], aliases=[], exclude=[], **kwargs):

        self.release = release
        self._groups = groups
        self._groups.set_parent(self)
        self.aliases = aliases
        self._exclude = exclude
        self.dap_datamodel = kwargs.get('dapdm', None)
        self.bitmasks = get_maskbits(self.release)
        self._mode = kwargs.get('mode', config.mode)
        self._keys = []
        self._get_parameters()
        self._check_datamodels()

    def __repr__(self):

        return ('<QueryDataModel release={0!r}, n_groups={1}, n_parameters={2}, n_total={3}>'
                .format(self.release, len(self.groups), len(self.parameters), len(self._keys)))

    def copy(self):
        """Returns a copy of the datamodel."""

        return copy_mod.deepcopy(self)

    def _get_parameters(self):
        ''' Get the parameters for the datamodel '''

        # if the release isn't in the allowed releases then do nothing
        # if self.release not in config._allowed_releases:
        #     self._groups = ParameterGroupList([])
        #     return

        # check cache for parameters
        if self.release in PARAM_CACHE:
            self._keys = PARAM_CACHE[self.release]
            return

        # get the parameters
        if self._mode == 'local':
            self._marvinform = MarvinForm(release=self.release)
            self._cleanup_keys()
        elif self._mode == 'remote':
            self._get_from_remote()
        elif self._mode == 'auto':
            if config.db:
                self._mode = 'local'
            else:
                self._mode = 'remote'
            self._get_parameters()

    def _get_from_remote(self):
        ''' Get the keys from a remote source '''

        from marvin.api.api import Interaction
        from brain import bconfig

        # if not a supported release, don't try to get params
        if not is_supported_release(self.release):
            return

        # if not urlmap then exit
        if not config.urlmap:
            self._keys = []
            return

        # try to get the url
        try:
            url = config.urlmap['api']['getparams']['url']
        except Exception as e:
            warnings.warn('Cannot access Marvin API to get the full list of query parameters. '
                          'for the Query Datamodel. Only showing the best ones.', MarvinUserWarning)
            url = None
            self._keys = []
            #self._cleanup_keys()

        # make the call
        if url:
            try:
                ii = Interaction(url, params={'release': self.release, 'paramdisplay': 'all'}, base=bconfig._collab_api_url)
            except Exception as e:
                warnings.warn('Could not remotely retrieve full set of parameters. {0}'.format(e), MarvinUserWarning)
                self._keys = []
            else:
                # this deals with all parameters from all releases at once
                # PARAM_CACHE.update(ii.getData())
                # self._check_aliases()

                # for key in list(PARAM_CACHE.keys()):
                #     keys = PARAM_CACHE[key] if key in PARAM_CACHE else []
                #     self._remove_query_params(keys=keys)

                # this deals with parameters per release
                self._keys = ii.getData()
                PARAM_CACHE[self.release] = self._keys
                self._check_aliases()
                self._remove_query_params()


    def _check_aliases(self):
        ''' Check the release of the return parameters against the aliases '''
        for key, val in PARAM_CACHE.items():
            if key != self.release and key in self.aliases:
                PARAM_CACHE[self.release] = val

    def _remove_query_params(self, keys=None):
        ''' Remove keys from query_params best list '''
        keys = keys or self._keys
        origlist = query_params.list_params('full')
        for okey in origlist:
            if okey in keys:
                keys.remove(okey)

        # add the final list to the cache
        self._keys = keys
        if self.release not in PARAM_CACHE:
            PARAM_CACHE[self.release] = keys

    def _cleanup_keys(self):
        ''' Cleans up the list for MarvinForm keys '''

        # get all the keys in the marvin form
        keys = list(self._marvinform._param_form_lookup.keys())
        keys.sort()

        # simplify the spaxelprop list down to one set
        mykeys = [k.split('.', 1)[-1] for k in keys if 'cleanspaxel' not in k]
        mykeys = [k.replace(k.split('.')[0], 'spaxelprop') if 'spaxelprop'
                  in k else k for k in mykeys]

        # replace table names with shortcut names
        rev = {v: k for k, v in self._marvinform._param_form_lookup._tableShortcuts.items()}
        newkeys = [k.replace(k.split('.')[0],
                             rev[k.split('.')[0]]) if k.split('.')[0] in rev.keys()
                   else k for k in mykeys]

        # remove any hidden keys
        newkeys = [n for n in newkeys if '._' not in n]

        # exclude tables from list of keys
        if self._exclude:
            for table in self._exclude:
                newkeys = [k for k in newkeys if table not in k]

        # final sort and set
        newkeys.sort()
        self._keys = newkeys

        # remove keys from query_params best list here
        self._remove_query_params()

    def _check_datamodels(self, parameters=None):
        ''' Check and match the datamodels '''

        params = [parameters] if parameters else self.parameters

        # DAP datamodel
        for qp in params:
            if self.dap_datamodel:
                try:
                    qp.property = self.dap_datamodel[qp.full]
                except ValueError as e:
                    pass

    def __eq__(self, value):
        """Uses fuzzywuzzy to return the closest parameter/group match."""

        # Gets the best match for parameters and groups. If there is a match
        # in parameters, returns it. Otherwise tries groups.

        try:
            param_best_match = self.parameters[value]
            if param_best_match:
                return param_best_match
        except (KeyError, ValueError):
            pass

        try:
            group_best_match = self.groups[value]
            if group_best_match:
                return group_best_match
        except (KeyError, ValueError):
            pass

        raise ValueError('too ambiguous input {!r}'.format(value))

    def __contains__(self, value):

        try:
            match = self.__eq__(value)
            if match is None:
                return False
            else:
                return True
        except ValueError:
            return False

    def __getitem__(self, value):
        return self == value

    @property
    def groups(self):
        """Returns the groups for this datamodel. """
        return self._groups

    @groups.setter
    def groups(self, value):
        """Raises an error if trying to set groups directly."""

        raise MarvinError('cannot set groups directly. Use add_groups() instead.')

    def add_group(self, group, copy=True):
        ''' '''
        self._groups.add_group(group, copy=copy, parent=self)

    def add_groups(self, groups, copy=True):
        ''' '''
        self._groups.add_groups(groups, copy=copy, parent=self)

    @property
    def parameters(self):
        return QueryList(sum([g.parameters for g in self.groups], []))

    def add_to_group(self, group, value=None):
        ''' Add free-floating Parameters into a Group '''

        # This helps in case that a group does not exist in the parameters,
        # for instance because we loaded the metadata query params.
        try:
            thegroup = self._groups == group
        except ValueError:
            return

        keys = []
        allkeys = copy_mod.copy(self._keys)
        if value is None:
            allkeys = []
            keys.extend(self._keys)
        else:
            for k in self._keys:
                if value in k:
                    mykey = allkeys.pop(allkeys.index(k))
                    if k not in thegroup.full:
                        keys.append(mykey)
        self._keys = allkeys
        thegroup.add_parameters(keys)

    def regroup(self, values):
        ''' Regroup a set of parameters into groups '''
        for group, value in values.items():
            if isinstance(value, list):
                for v in value:
                    self.add_to_group(group, v)
            else:
                self.add_to_group(group, value)

    @property
    def best(self):
        return QueryList([p for p in self.parameters if p.best is True])

    @property
    def best_groups(self):
        return self.groups.best

    def set_best(self, best):
        ''' sets a list of best query parameters '''
        for b in best:
            if isinstance(b, QueryParameter):
                b.best = True
            elif isinstance(b, six.string_types):
                for i, p in enumerate(self.parameters):
                    if b == p.full:
                        self.parameters[i].best = True

    def use_all_spaxels(self):
        ''' Sets the datamodel to use all the spaxels '''
        self._marvinform = MarvinForm(release=self.release, allspaxels=True)
        self._cleanup_keys()

    def _reset_marvinform(self):
        self._marvinform = MarvinForm(release=self.release)

    def to_table(self, pprint=False, max_width=1000, only_best=False, db=False):
        ''' Write the datamodel to an Astropy table '''

        param_table = Table(None, names=['group', 'full', 'best', 'name', 'display',
                                         'db_schema', 'db_table', 'db_column', 'is_hybrid'],
                            dtype=['S20', 'S300', bool, 'S30', 'S30', 'S30', 'S30', 'S30', bool])

        iterable = self.groups.parameters

        param_table.meta['release'] = self.release

        for param in iterable:
                param_table.add_row((param.group, param.full, param.best,
                                     param.name, param.display,
                                     param.db_schema, param.db_table,
                                     param.db_column, param.is_hybrid()))

        if only_best:
            notbest = param_table['best'] is False
            param_table.remove_rows(notbest)

        if not db:
            param_table.remove_columns(['db_schema', 'db_table', 'db_column', 'is_hybrid'])

        if pprint:
            param_table.pprint(max_width=max_width, max_lines=1e6)
            return

        return param_table

    def write_csv(self, filename=None, path=None, overwrite=None, **kwargs):
        ''' Write the datamodel to a CSV '''

        release = self.release.lower().replace('-', '')

        if not filename:
            filename = 'query_dm_{0}.csv'.format(release)

        if not path:
            path = os.path.join(os.getenv("MARVIN_DIR"), 'docs', 'sphinx', '_static')

        fullpath = os.path.join(path, filename)
        table = self.to_table(**kwargs)
        table.write(fullpath, format='csv', overwrite=overwrite)


class QueryDataModelList(DataModelList):
    """A dictionary of Query datamodels."""
    base = {'QueryDataModel': QueryDataModel}


def get_best_fuzzy(name, choices, cutoff=60, return_score=False):
    items = process.extractBests(name, choices, score_cutoff=cutoff, scorer=fuzz.WRatio)

    if not items:
        best = None
    elif len(items) == 1:
        best = items[0]
    else:
        scores = [s[1] for s in items]
        # finds items with the same score
        morethanone = sum(np.max(scores) == scores) > 1
        if morethanone:
            # tries to find an exact string match
            exact = []
            for s in items:
                itemname = s[0].name if isinstance(s[0], QueryParameter) else s[0]
                if itemname.lower() == name.lower():
                    exact.append(s)
            # returns exact match or fails with ambiguity
            if exact:
                best = exact[0]
            else:
                options = [s[0].name if isinstance(s[0], QueryParameter)
                           else s[0] for s in items if s[1] == np.max(scores)]
                raise KeyError('{0} is too ambiguous.  '
                               'Did you mean one of {1}?'.format(name, options))
        else:
            best = items[0]

    if best is None:
        raise ValueError('Could not find a match for {0}.  Please refine your text.'.format(name))

    return best if return_score else best[0]


def strip_mapped(self):
    ''' Strip the mapped items for display with dir

    Since __dir__ cannot have . in the attribute name, this
    strips the returned mapper(item) parameter of any . in the name.
    Used for query parameter syntax [table.parameter_name]

    For cases where the parameter_name is "name", and thus
    non-unique, it also returns the mapper name with "." replaced
    with "_", to make unique. "ifu.name" becomes "ifu_name", etc.

    Parameters:
        self:
            a QueryFuzzyList object

    Returns:
        list of mapped named stripped of dots

    '''
    params = []
    for item in self:
        mapped_item = self.mapper(item)
        if '.' in mapped_item:
            parta, name = mapped_item.split('.')
            if name == 'name':
                name = mapped_item.replace('.', '_')
            params.append(name)
        else:
            params.append(mapped_item)
    return params


class QueryFuzzyList(FuzzyList):
    ''' Fuzzy List for Query Parameters '''

    def __dir__(self):
        class_members = list(list(zip(*inspect.getmembers(self.__class__)))[0])
        params = strip_mapped(self)
        return class_members + params

    def __getattr__(self, value):

        mapped_values = [super(QueryFuzzyList, self).__getattribute__('mapper')(item)
                         for item in self]
        stripped_values = strip_mapped(self)

        if value in stripped_values:
            return self[mapped_values[stripped_values.index(value)]]

        return super(QueryFuzzyList, self).__getattribute__(value)

    def index(self, value):
        param = self == value
        index = [i for i, item in enumerate(self) if item is param]

        if index:
            return index[0]
        else:
            raise ValueError('{0} is not in the list'.format(value))

    def pop(self, value=None):
        if isinstance(value, int):
            return super(QueryFuzzyList, self).pop(value)
        elif not value:
            return super(QueryFuzzyList, self).pop()
        else:
            idx = self.index(value)
            return super(QueryFuzzyList, self).pop(idx)

    def remove(self, value):
        param = value == self
        tmp = [n for n in self if n != param]
        QueryFuzzyList.__init__(self, tmp, use_fuzzy=get_best_fuzzy)


class ParameterGroupList(QueryFuzzyList):
    ''' ParameterGroup Object

    This object inherits from the Python list object. This
    represents a list of query ParameterGroups.

    '''

    def __init__(self, items):
        self.score = None
        paramgroups = self._make_groups(items)
        QueryFuzzyList.__init__(self, paramgroups, use_fuzzy=get_best_fuzzy)

    def mapper(self, value):
        return value.name.lower().replace(' ', '_').replace('.', '_')

    def _make_groups(self, items, best=None):
        if isinstance(items, list):
            paramgroups = [ParameterGroup(item, []) for item in items]
        elif isinstance(items, dict):
            paramgroups = [ParameterGroup(key, vals)
                           for key, vals in items.items()]
        elif isinstance(items, six.string_types):
            paramgroups = ParameterGroup(items, [])
        return paramgroups

    def set_parent(self, parent):
        """Sets parent."""

        assert isinstance(parent, QueryDataModel), 'parent must be a QueryDataModel'
        self.parent = parent
        for item in self:
            item.set_parent(parent)

    @property
    def parameters(self):
        ''' List all the queryable parameters '''
        return QueryList([param for group in self for param in group])

    @property
    def names(self):
        ''' List all the parameter groups '''
        return self.list_groups()

    @property
    def best(self):
        ''' List the best parameters in each group '''
        grp_copy = copy_mod.deepcopy(self)
        grp_copy.__init__(bestparams)
        grp_copy.parent._check_datamodels()
        return grp_copy

    @names.setter
    def names(self, value):
        """Raises an error if trying to set groups directly."""

        raise MarvinError('cannot set names directly. Use add_groups() instead.')

    def list_groups(self):
        '''Returns a list of query groups.

        Returns:
            names (list):
                A string list of all the Query Group names
        '''
        return [group.name for group in self]

    def add_group(self, group, copy=True, parent=None):
        ''' '''

        # don't add the group if already in there
        name = str(group)
        if name in self:
            return

        new_grp = copy_mod.copy(group) if copy else group
        if isinstance(new_grp, ParameterGroup):
            self.append(new_grp)
        else:
            new_grp = self._make_groups(new_grp)
            new_grp.set_parent(parent)
            self.append(new_grp)

    def add_groups(self, groups, copy=True, parent=None):
        ''' '''
        for group in groups:
            self.add_group(group, copy=copy, parent=parent)

    def list_params(self, name_type='full', groups=None):
        '''Returns a list of parameters from all groups.

        Return a string list of the full parameter names.
        Default is all parameters across all groups.

        Parameters:
            groups (str|list):
                A string or list of strings representing the groups
                of parameters you wish to return
            name_type (str):
                The type of name to generate (full, name, short, remote, display).
                Default is full.

        Returns:
            params (list):
                A list of full parameter names
        '''

        assert name_type in ['full', 'short', 'name', 'remote', 'display'], \
            'name_type must be (full, short, name, remote, display)'

        if groups:
            groups = groups if isinstance(groups, list) else [groups]
            grouplist = [self[g] for g in groups]
            return [param.__getattribute__(name_type) for group in grouplist for param in group]
        else:
            return [param.__getattribute__(name_type) for group in self for param in group]


class ParameterGroup(QueryFuzzyList):
    ''' A Query Parameter Group Object

    Query parameters are grouped into specific categories
    for ease of use and navigation.  This object subclasses
    from the Python list object.

    Parameters:
        name (str):
            The name of the group
        items (list|dict):
            A list or dictionary of parameters.  If a list of names is input,
            each name will be used as the full name.

    '''
    def __init__(self, name, items, parent=None):
        self.name = name
        self.score = None
        self.parent = parent

        queryparams = []
        for item in items:
            this_param = self._make_query_parameter(item)
            queryparams.append(this_param)

        QueryFuzzyList.__init__(self, queryparams, use_fuzzy=get_best_fuzzy)
        self._check_names()
        if self.parent:
            self._check_datamodels()

    def __repr__(self):
        return '<ParameterGroup name={0.name}, n_parameters={1}>'.format(self, len(self))

    def __str__(self):
        return self.name

    def mapper(self, value):
        return value.full

    def _make_query_parameter(self, item):
        ''' Create and return a QueryParameter '''
        if isinstance(item, dict):
            item.update({'group': self.name, 'best': True})
            this_param = QueryParameter(**item)
        elif isinstance(item, six.string_types):
            is_best = [p for p in query_params.parameters
                       if item in p.full and (item == p.name or item == p.full)]
            if is_best:
                best_dict = is_best[0].__dict__
                best_dict.update({'group': self.name, 'best': True})
                this_param = QueryParameter(**best_dict)
            else:
                this_param = QueryParameter(item, group=self.name, best=False)
        if self.parent:
            this_param.set_parents(self.parent, self)

        return this_param

    def set_parent(self, parent):
        """Sets datamodel parent."""

        assert isinstance(parent, QueryDataModel), 'parent must be a QueryDataModel'
        self.parent = parent
        for item in self:
            item.set_parents(parent, self)

    @property
    def full(self):
        return self.list_params('full')

    @property
    def remote(self):
        return self.list_params('remote')

    @property
    def short(self):
        return self.list_params('short')

    @property
    def display(self):
        return self.list_params('display')

    @property
    def parameters(self):
        return self.list_params()

    @parameters.setter
    def parameters(self, value):
        """Raises an error if trying to set groups directly."""

        raise MarvinError('cannot set groups directly. Use add_parameters() instead.')

    def add_parameter(self, value, copy=True):
        ''' '''

        new_par = copy_mod.copy(value) if copy else value
        if isinstance(value, QueryParameter):
            self.append(new_par)
        else:
            new_qp = self._make_query_parameter(new_par)
            self.append(new_qp)

    def add_parameters(self, values, copy=True):
        ''' '''
        for value in values:
            self.add_parameter(value, copy=copy)

    def list_params(self, name_type=None, subset=None):
        ''' List the parameter names for a given group

        Lists the Query Parameters of the given group

        Parameters:
            subset (str|list):
                String list of a subset of parameters to return
            name_type (str):
                The type of name to generate (full, short, name, remote, display).

        Returns:
            param (list):
                The list of parameter
        '''

        assert name_type in ['full', 'short', 'name', 'remote', 'display', None], \
            'name_type must be (full, short, name, remote, display)'

        if subset:
            params = subset if isinstance(subset, list) else [subset]
            paramlist = [self[g] for g in params]
        else:
            paramlist = self

        if name_type:
            return [param.__getattribute__(name_type) for param in paramlist]
        else:
            return QueryList([param for param in paramlist])

    def _check_names(self):
        names = self.list_params('remote')
        for i, name in enumerate(names):
            if names.count(name) > 1:
                self[i].remote = self[i]._under
                self[i].name = self[i]._under
                self[i].short = self[i]._under
                self[i].display = '{0} {1}'.format(self[i].table.title(), self[i].display)

    def _check_datamodels(self):
        for item in self:
            if not item.property:
                if item.full in self.parent.dap_datamodel:
                    item.property = self.parent.dap_datamodel[item.full]


class QueryList(QueryFuzzyList):
    ''' A class for a list of Query Parameters '''

    def __init__(self, items):
        QueryFuzzyList.__init__(self, items, use_fuzzy=get_best_fuzzy)
        self._full = [s.full for s in self]
        self._remote = [s.remote for s in self]
        self._short = [s.short for s in self]

    def mapper(self, value):
        return value.full

    def get_full_from_remote(self, value):
        ''' Get the full name from the remote name '''
        return self._full[self._remote.index(value)]


class QueryParameter(object):
    ''' A Query Parameter class

    An object representing a query parameter.  Provides access to
    different names for a given parameter.

    Parameters:
        full (str):
            The full naming syntax (table.name) used for all queries.
            This name is recommended for full uniqueness.
        table (str):
            The name of the database table the parameter belongs to
        name (str):
            The name of the parameter in the database
        short (str):
            A shorthand name of the parameter
        display (str):
            A display name used for web and plotting purposes.
        dtype (str):
            The type of the parameter (e.g. string, integer, float)

    Attributes:
        property:
            The DAP Datamodel Property corresponding to this Query Parameter
    '''

    def __init__(self, full, table=None, name=None, short=None, display=None,
                 remote=None, dtype=None, **kwargs):
        self.full = full
        self.table = table
        self.name = name
        self.short = short
        self.display = display
        self.dtype = dtype
        self.remote = remote
        self.value = kwargs.get('value', None)
        self.group = kwargs.get('group', None)
        self.schema = kwargs.get('schema', None)
        self.best = kwargs.get('best', None)
        self.property = None
        self.parent = None
        self._set_names()

    def __repr__(self):
        return ('<QueryParameter full={0.full}, name={0.name}, short={0.short}, '
                'remote={0.remote}, display={0.display}>'.format(self))

    def __str__(self):
        return self.full

    def _split_full(self):
        ''' Split the full name into parts '''
        schema = table = column = None
        if self.full.count('.') == 1:
            table, column = self.full.split('.')
        elif self.full.count('.') == 2:
            schema, table, column = self.full.split('.')
        else:
            column = self.full

        return schema, table, column

    def _set_names(self):
        ''' Sets alternate names if it can '''
        self._under = self.full.replace('.', '_')
        if not self.table or not self.name:
            if self.full.count('.') == 1:
                self.table, self.name = self.full.split('.')
            elif self.full.count('.') == 2:
                self.schema, self.table, self.name = self.full.split('.')
                self.full = '{0}.{1}'.format(self.table, self.name)
            else:
                self.name = self.full
        if not self.short:
            self.short = self._under if 'name' == self.name else self.name
        if not self.display:
            self.display = self.name.title()
        if not self.remote:
            self.remote = self._under if 'name' == self.name else self.name
        # correct the name if just name
        self.name = self._under if 'name' == self.name else self.name
        # used for a token string on the web
        self._joinedname = ', '.join([self.full, self.name, self.short, self.display])

    def set_parents(self, parent, group):
        ''' Set the parent datamodel '''

        assert isinstance(parent, QueryDataModel), 'parent must be a QueryDataModel'
        assert isinstance(group, ParameterGroup), 'group must be a ParameterGroup'
        self.parent = parent
        self.parent_group = group

    @property
    def db_schema(self):
        schema, table, column = self._split_full()
        if not schema:
            if self._in_form():
                schema = self.parent._marvinform._param_form_lookup[self.full].Meta.model.__table__.schema
            else:
                schema = None
        return schema

    @property
    def db_table(self):
        schema, table, column = self._split_full()
        release_num = self.parent.release.split('-')[1]
        table = table + release_num if table == 'spaxelprop' else table
        return table

    @property
    def db_column(self):
        schema, table, column = self._split_full()
        return column

    def is_hybrid(self):
        if self._in_form():
            model = self.parent._marvinform._param_form_lookup[self.full].Meta.model
            hybrids = get_hybrid_properties(model).keys()
            return self.db_column in hybrids
        return None

    def _in_form(self):
        ''' Check in parameters is in the Marvin Form '''
        if not hasattr(self.parent, '_marvinform'):
            return False

        return self.full in self.parent._marvinform._param_form_lookup


# # Get the Common Parameters from the filelist

def get_params():

    bestpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../../../data', 'query_params_best.cfg')

    if os.path.isfile(bestpath):
        with open(bestpath, 'r') as stream:
            bestparams = yaml.load(stream, Loader=yamlordereddictloader.Loader)

        # remove DAP spaxel properties from the list
        #if config.access not in config._dap_query_modes:
        if not config._allow_DAP_queries:
            b = bestparams.copy()
            for grp, props in b.items():
                props = [prop for prop in props if prop['table'] != 'spaxelprop']
                bestparams[grp] = props
                if not props:
                    bestparams.pop(grp)

        return bestparams
    else:
        return None


bestparams = get_params()
query_params = ParameterGroupList(bestparams)



#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-04-28 14:07:58
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-04-28 14:07:58 by Brian Cherinka
    Last Modified On: 2016-04-28 14:07:58 by Brian

'''
from __future__ import print_function
from __future__ import division

from flask import Blueprint, render_template, request
from flask_classful import route
from marvin.core.exceptions import MarvinError
from marvin.tools.plate import Plate as mPlate
from marvin.utils.general import getImagesByPlate
from marvin.web.controllers import BaseWebView
from marvin.web.web_utils import buildImageDict
from marvin.api.base import arg_validate as av
from marvin.web.extensions import cache

plate = Blueprint("plate_page", __name__)


class Plate(BaseWebView):
    route_base = '/plate/'

    def __init__(self):
        ''' Initialize the route '''
        super(Plate, self).__init__('marvin-plate')
        self.plate = self.base.copy()
        self.plate['plateid'] = None

    def before_request(self, *args, **kwargs):
        ''' Do these things before a request to any route '''
        super(Plate, self).before_request(*args, **kwargs)
        self.reset_dict(self.plate)

    @route('/', methods=['GET', 'POST'])
    def index(self):

        return render_template('plate.html', **self.plate)

    @cache.memoize()
    def get(self, plateid):
        ''' Retrieve info for a given plate id '''

        # validate the input
        args = av.manual_parse(self, request)

        self.plate['plateid'] = args.get('plateid')
        pinputs = {'plate': plateid, 'mode': 'local', 'nocubes': True, 'release': self._release}
        try:
            plate = mPlate(**pinputs)
        except MarvinError as e:
            self.plate['plate'] = None
            self.plate['drpver'] = self._drpver
            self.plate['error'] = 'Could not grab Plate for id {0}: {1}'.format(plateid, e)
        else:
            self.plate['plate'] = plate
            self.plate['drpver'] = plate._drpver
            tmpfile = plate._getFullPath(url=True)
            self.plate['sasurl'] = plate.platedir

        # Get images for plate
        imfiles = None
        try:
            imfiles = getImagesByPlate(plateid=plateid, as_url=True, mode='local', release=self._release)
        except MarvinError as e:
            self.plate['error'] = 'Error: could not get images for plate {0}: {1}'.format(plateid, e)
        else:
            # thumbs = [imfiles.pop(imfiles.index(t)) if 'thumb' in t else t for t in imfiles]
            # plateifu = ['-'.join(re.findall('\d{3,5}', im)) for im in imfiles]
            images = buildImageDict(imfiles)

        # if image grab failed, make placeholders
        if not imfiles:
            images = buildImageDict(imfiles, test=True, num=29)

        # Add images to dict
        self.plate['images'] = images

        return render_template('plate.html', **self.plate)


def get_plate_cache(*args, **kwargs):
    ''' Function used to generate the route cache key

    Cache key when using cache.memoize or cache.cached decorator.
    memoize remembers input methods arguments; cached does not.

    Parameters:
        args (list):
            a list of the fx/method route call and object instance (self)
        kwargs (dict):
            a dictonary of arguments passed into the method
    Returns:
        A string used for the cache key lookup
    '''
    # get the plateid from the keyword argument
    plateid = kwargs.get('plateid')
    # get the Plate method and self instance
    fxn, inst = args

    # create unique cache key name
    key = 'getplate_{0}_{1}'.format(plateid, inst._release.lower().replace('-', ''))
    # append if logged in
    if inst.plate['loggedin']:
        key = '{0}_loggedin'.format(key)
    return key


Plate.get.make_cache_key = get_plate_cache

Plate.register(plate)



from __future__ import print_function, division, absolute_import

from marvin.tools import maps

from .base import VACMixIn, VACTarget

import numpy as np
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib as mpl
import scipy.linalg as sl
import json
import marvin.utils.dap.bpt as bpt
import marvin

from astropy.io import fits
from astropy.wcs import WCS
from astropy.table import Table
from scipy.interpolate import RectBivariateSpline
from scipy.spatial import distance_matrix
from marvin.tools.quantities.spectrum import Spectrum
from marvin import log

LUT = {7: 3, 19: 5, 37: 7, 61: 9, 91: 11, 127: 13}
spaxel_grid = {7: 24, 19: 34, 37: 44, 61: 54, 91: 64, 127: 74}


def convert_json(table, column_name):
    '''Unpacks the JSON column of a table

    Paramters:
        table (astropy.table.Table):
            An astropy table
        column_name (str):
            The name of the column made up of JSON strings

    The input table is updated in place by appending `_string` to
    the end of the JSON column name and adding a new column with
    `_list` on the end with the list representation of the same
    column.
    '''
    # this unpacks the json column of a table
    new_col = [json.loads(i) for i in table[column_name]]
    table.rename_column(column_name, '{0}_string'.format(column_name))
    table['{0}_list'.format(column_name)] = new_col


def non_blank(table, *column_name):
    '''Count how many non-blank classifications are in the given columns
    of the input table.

    Paramters:
        table (astropy.table.Table):
            An astropy table with Zooniverse classifications
        column_name(s) (str):
            One or multiple column names

    Returns:
        non_blank (int):
            The total number of non-blank classifications across *all*
            input column names (combined with "logical or" in a single
            row).
    '''
    for cdx, c in enumerate(column_name):
        if cdx == 0:
            non_blank = np.array([len(i) > 0 for i in table[c]])
        else:
            non_blank = non_blank | np.array([len(i) > 0 for i in table[c]])
    return non_blank.sum()


def cov_to_ellipse(cov, pos, nstd=1, **kwargs):
    '''Create a covariance ellipse given an covariance matrix and postion

    Paramters:
        cov (numpy.array):
            2x2 covariance matrix
        pos (numpy.array):
            1x2 center position of the ellipse

    Keywords:
        nstd (int):
            Number of standard deviations to make the output ellipse (Default=1)
        kwargs:
            All other keywords are passed to matplotlib.patches.Ellipse

    Returns:
        ellipse (matplotlib.patches.Ellipse):
            matplotlib ellipse patch object
    '''
    eigvec, eigval, V = sl.svd(cov, full_matrices=False)
    # the angle the first eigenvector makes with the x-axis
    theta = np.degrees(np.arctan2(eigvec[1, 0], eigvec[0, 0]))
    # full width and height of ellipse, not radius
    # the eigenvalues are the variance along the eigenvectors
    width, height = 2 * nstd * np.sqrt(eigval)
    return patches.Ellipse(xy=pos, width=width, height=height, angle=theta, **kwargs)


def alpha_overlay(C_a, a_a, C_b, a_b=None):
    '''Take a base color (C_a), an alpha map (a_a), background image (C_b), and optional
    background alpha map (a_b) and overlay them.

    Paramters:
        C_a (numpy.array):
            1x3 RGB array for the base color to be overlayed
        a_a (numpy.array):
            NxM array of alpha values for each postion on an image
        C_b (numpy.array):
            1x3 RGB array for the background color or NxMx3 RGB array
            for a background image
        a_b (numpy.array):
            NxM array of alpha values for the background color/image

    Returns:
        c_out (numpy.array):
            NxMx3 RGB array containing the alpha overlayed image.
    '''
    if a_b is None:
        a_b = np.ones(a_a.shape)
    c_a = np.array([a_a.T] * 3).T * C_a
    c_b = np.array([a_b.T] * 3).T * C_b
    c_out = c_a + ((1 - a_a.T) * c_b.T).T
    return c_out


def alpha_maps(maps, colors=None, vmin=0, vmax=15, background_image=None):
    '''Take a list of color masks and base color values
    and make an alpha-mask overlay image.

    Parameters:
        maps (list):
            List of masks to use as alpha maps

    Keywords:
        colors (list):
            What matplotlib color to use for each of the input maps
            (defaults to standard MPL color cycle)
        vmin (int):
            Value in the maps at or below this value will be 100% transparent
        vmax (int):
            Value in the maps at or above this value will be 100% opaque
        background_image (numpy.array):
            RGB array to use as the background image (default solid white)

    Returns:
        overlay (numpy.array):
            RGB array with each map overlayed on each other with alpha
            transparency.
    '''
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
    iter_cycle = iter(mpl.rcParams['axes.prop_cycle'])
    for mdx, m in enumerate(maps):
        if colors is None:
            c = next(iter_cycle)['color']
        else:
            c = colors[mdx]
        base_color = np.array(mpl.colors.to_rgb(c))
        norm_map = norm(m)
        if mdx == 0:
            if background_image is None:
                background_color = np.ones(3)
            else:
                background_color = background_image
        background_color = alpha_overlay(base_color, norm_map, background_color)
    return background_color


def make_alpha_bar(color, vmin=-1, vmax=15):
    '''Make a matplotlib color bar for a alpha mask of a single color

    Parameters:
        color (string):
            A matplotlib color (any format matplotlib accepts)

    Keywords:
        vmin (int):
            The minimum value for the colorbar. Default value is -1
            to ensure the labels show up correctly when used with
            plot_alpha_bar.
        vmax (int):
            The maximum value for the colorbar. Default value is 15.

    Returns:
        colormap (mpl.colors.ListedColormap):
            The colormap for the colorbar
        norm (mpl.colors.Normalize):
            The normalization for the color bar
    '''
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
    a_a = norm(range(vmin, vmax))
    C_a = np.array(mpl.colors.to_rgb(color))
    new_cm = alpha_overlay(C_a, a_a, np.ones(3))
    return mpl.colors.ListedColormap(new_cm), norm


def make_alpha_color(count, color, vmin=1, vmax=15):
    '''Give a matplotlib color and alpha channel proportional to
    the input count value.

    Parameters:
        count (int):
            The count value used to select an alpha value
        color (string):
            A matplotlib color (any format matplotlib accepts)

    Keywords:
        vmin (int):
            The count value to be associated with transparent.
            Default is 1.
        vmax (int):
            The count value to be associated with opaque. Default
            is 15.

    Returns:
        alpha_color (tuple):
            An rgba tuple for the new alpha color
    '''
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
    return mpl.colors.to_rgb(color) + (norm(count), )


def plot_alpha_bar(color, grid, ticks=[]):
    '''Display and alpha colorbar on a plot grid.

    Parameters:
        color (string):
            A matplotlib color (any format matplotlib accepts)
        grid (matplotlib.gridspec.SubplotSpec):
            A gridspec subplot specification to place the color bar in

    Keywords:
        ticks (list):
            A list of tick value for the colorbar

    Returns:
        ax_bar (matplotlib.axes.Axes):
            Matplotlib axis object for the colorbar
        colorbar (mpl.colorbar.ColorbarBase):
            Matplotlib colorbar object
    '''
    bar, norm = make_alpha_bar(color)
    ax_bar = plt.subplot(grid)
    cb = mpl.colorbar.ColorbarBase(ax_bar, cmap=bar, norm=norm, orientation='vertical', ticks=ticks)
    cb.outline.set_linewidth(0)
    return ax_bar, cb


def plot_alpha_scatter(x, y, mask, color, ax, snr=None, sf_mask=None, value=True, **kwargs):
    '''Make a scatter plot where each x-y point has and alpha transparency
    set by the values in a count mask array.

    Parmeters:
        x (numpy.array):
            1-D numpy array with x-values to be plotted
        y (numpy.array or spectral line object from a Marvin Maps cube):
            1-D numpy array with y-values to be plotted
        mask (numpy.array):
            1-D numpy array with mask array containing the "count" value for each
            (x,y) data point
        color (string):
            A matplotlib color (any format matplotlib accepts) used for the
            base color of the data points
        ax (matplotlib.axes.Axes):
            The maplotlib axes to use for the plot

    Keywords:
        snr (float):
            Minimum signal to noise ratio to use as a cutoff for the y-values.
            Defaults to `None`. Only used if `value=True`
        sf_mask (numpy.array):
            1-D numpy array with A star formation region mask that is 1 when there
            is star formation in a spaxel and 0 otherwise. If passed in only spxels
            where this mask is 1 will be plotted.
        value (bool):
            If True y is a spectral line object from a Marvin Maps cube, otherwise
            y is assumed to be regular np.array object.
        **kwargs:
            All other keywords are passed forward to matplotlib's scatter plot
            function.

    Returns:
        scatter (matplotlib.collections.PathCollection):
            A maplotlib scatter plot object
    '''
    idx = mask > 0
    if value:
        idx = idx & (y.value > 0)
    if (value) and (snr is not None):
        idx = idx & (y.snr > snr)
    if sf_mask is not None:
        idx = idx & sf_mask
    c = mpl.colors.to_rgb(color)
    c_a = np.array([c + (i, ) for i in mask[idx] / 15])
    c_a[c_a > 1] = 1
    if idx.sum() > 0:
        if value:
            return ax.scatter(x[idx], y.value[idx], c=c_a, edgecolor=c_a, **kwargs)
        else:
            return ax.scatter(x[idx], y[idx], c=c_a, edgecolor=c_a, **kwargs)
    return None


class GZ3DVAC(VACMixIn):
    '''Provides access to the Galaxy Zoo 3D spaxel masks.

    VAC name: Galaxy Zoo: 3D

    URL: https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=galaxy-zoo-3d

    Description: Galaxy Zoo: 3D (GZ: 3D) made use of a project on the Zooniverse platform to
        crowdsource spaxel masks locating galaxy centers, foreground stars, bars and spirals
        in the SDSS images of MaNGA target galaxies. These masks (available for use within Marvin)
        can be used to pick out spectra, or map quantities associated with the different
        structures. See Masters et al. 2021 for more information, advice on useage and examples.

    Authors: Coleman Krawczyk, Karen Masters and the rest of the Galaxy Zoo 3D Team.
    '''

    name = 'gz3d'
    description = 'Return object for working with Galaxy Zoo 3D data masks'
    version = {'DR17': 'v4_0_0', 'MPL-11': 'v4_0_0'}
    display_name = 'Galaxy Zoo 3D'
    url = 'https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=galaxy-zoo-3d'
    
    include = (marvin.tools.cube.Cube, marvin.tools.maps.Maps)

    def set_summary_file(self, release):
        ''' Sets the path to the GalaxyZoo3D summary file '''
        self.path_params = {'ver': self.version[release]}

        self.summary_file = self.get_path('mangagz3dmetadata', path_params=self.path_params)
        self.center_summary_file = self.get_path('mangagz3dcenters', path_params=self.path_params)
        self.stars_summary_file = self.get_path('mangagz3dstars', path_params=self.path_params)

    def get_target(self, parent_object):
        '''Find the GZ3D data based on the manga ID'''
        mangaid = parent_object.mangaid

        if parent_object.__class__ == marvin.tools.cube.Cube:
            cube = parent_object
            maps = parent_object.getMaps()
        else:
            cube = parent_object.getCube()
            maps = parent_object

        if not self.file_exists(self.summary_file):
            self.summary_file = self.download_vac('mangagz3dmetadata', path_params=self.path_params)
            self.center_summary_file = self.download_vac('mangagz3dcenters', path_params=self.path_params)
            self.stars_summary_file = self.download_vac('mangagz3dstars', path_params=self.path_params)

        summary_table = Table.read(self.summary_file, hdu=1)
        # Table adds extra spaces to short strings, these need to be stripped off
        gz3d_mangaids = np.array([mid.strip() for mid in summary_table['MANGAID']])

        idx = gz3d_mangaids == mangaid
        if idx.sum() > 0:
            file_name = summary_table[idx]['file_name'][0].strip()

            self.path_params.update(file_name=file_name)
            self.gz3d_filename = self.get_path('mangagz3d', path_params=self.path_params)

            if not self.file_exists(self.gz3d_filename):
                self.gz3d_filename = self.download_vac('mangagz3d', path_params=self.path_params)

            return GZ3DTarget(self.gz3d_filename, cube, maps)

        log.info('There is no GZ3D data for this mangaid: {0}'.format(mangaid))
        return None


class GZ3DTarget(object):
    '''A customized class to open and display GZ3D spaxel masks

    Parameters:
        filename (str):
            Path to the GZ3D fits file
        cube (marvin.tools.cube.Cube):
            Marvin Cube object
        maps (marvin.tools.maps.Maps):
            Mavin Maps object

    Attributes:
        hdulist (list):
            List containing the 11 HDUs present in the GZ3D fits file (see <url> for full data model)
        wcs (astropy.wcs):
            WCS object for the GZ3D masks (e.g. HDU[1] to HDU[4])
        image (numpy.array):
            The galaxy image shown to GZ3D volunteers
        center_mask (numpy.array):
            Pixel mask (same shape as image) of the clustering results for the galaxy center(s).  Each identified
            center is represented by a 2 sigma ellipse of clustered points with the value of the pixels inside
            the ellipse equal to the number of points belonging to that cluster.
        star_mask (numpy.array):
            Pixel mask (same shape as image) of the clustering results for forground star(s).  Each identified
            star is represented by a 2 sigma ellipse of clustered points with the value of the pixels inside
            the ellipse equal to the number of points belonging to that cluster.
        spiral_mask (numpy.array):
            Pixel mask (same shape as image) of the spiral arm location(s).  The value for the pixels is the number
            of overlapping polygons at that location.
        bar_mask (numpy.array):
            Pixel mask (same shape as image) of the bar location.  The value for the pixels is the number of
            overlapping polygons at that location.
        metadata (astropy.Table):
            Table containing metadata about the galaxy.
        ifu_size (int):
            Size of IFU
        center_clusters (astropy.Table):
            Position for identified galaxy center(s) in both image pixels and (RA, DEC)
        num_centers (int):
            Number of galaxy centers identified
        star_clusters (astropy.Table):
            Position for identified forground star(s) in both image pixels and (RA, DEC)
        num_stars (int):
            Number of forground stars identified
        center_star_classifications (astropy.Table):
            Raw GZ3D classifications for center(s) and star(s)
        num_center_star_classifications (int):
            Total number of classifications made for either center(s) or star(s)
        num_center_star_classifications_non_blank (int):
            Total number of non-blank classifications made for either center(s) or star(s)
        spiral_classifications (astropy.Table):
            Raw GZ3D classifications for spiral arms
        num_spiral_classifications (int):
            Total number of spiral arm classifications made
        num_spiral_classifications_non_blank (int):
            Total number of non-blank spiral arm classifications made
        bar_classifications (astropy.Table):
            Raw GZ3D classifications for bars
        num_bar_classifications (int):
            Total number of bar classifications made
        num_bar_classifications_non_blank (int):
            Total number of non-blank bar classifications made
        cube (marvin.tools.cube.Cube):
            Marvin Cube object
        maps (marvin.tools.maps.Maps)
            Marvin Maps object
        center_mask_spaxel (numpy.array):
            The center_mask projected into spaxel space
        star_mask_spaxel (numpy.array):
            The star_mask projected into spaxel space
        spiral_mask_spaxel (numpy.array):
            The spiral_mask projected into spaxel space
        bar_mask_spaxel (numpy.array):
            The bar_mask projected into spaxel space
        other_mask_spaxel (numpy.array):
            A mask for spaxel not contained in any of the other spaxel masks
    '''
    def __init__(self, filename, cube, maps):
        '''Set useful paramters and process the GZ3D masks'''
        # get the subject id from the filename
        self.subject_id = filename.split('/')[-1].split('_')[-1].split('.')[0]
        self.cube = cube
        self.maps = maps
        self.mean_bar = None
        self.mean_spiral = None
        self.mean_center = None
        self.mean_not_bar = None
        self.mean_not_spiral = None
        self.mean_not_center = None
        self.log_oiii_hb = None
        self.log_nii_ha = None
        self.log_sii_ha = None
        self.log_oi_ha = None
        self.dis = None
        # read in the fits file

        with fits.open(filename) as hdulist:
            self.hdulist = hdulist
            # grab the wcs
            self.wcs = WCS(self.hdulist[1].header)
            self._process_images()
            # read in metadata
            self.metadata = Table(self.hdulist[5].data)
            self.ifu_size = int(self.metadata['IFUDESIGNSIZE'][0])
            self._process_clusters()
            self._process_clusters_classifications()
            self._process_spiral_classifications()
            self._process_bar_classifications()
            self._process_all_spaxel_masks()
            self._get_bpt()

    def _process_images(self):
        '''Extract the data from the fits file and give it useful names'''
        # read in images
        self.image = self.hdulist[0].data
        self.center_mask = self.hdulist[1].data
        self.star_mask = self.hdulist[2].data
        self.spiral_mask = self.hdulist[3].data
        self.bar_mask = self.hdulist[4].data

    def _process_clusters(self):
        '''Format the cluster tables form the fits file as Astropy tables and count the number of rows'''
        # read in center clusters
        self.center_clusters = Table(self.hdulist[6].data)
        self.num_centers = len(self.center_clusters)
        # read in star clusters
        self.star_clusters = Table(self.hdulist[7].data)
        self.num_stars = len(self.star_clusters)

    def _process_clusters_classifications(self):
        '''Format the Zooniverse point classifications as Astropy tables and count the number of unique non-blank classifications'''
        # read in center and star classifications
        self.center_star_classifications = Table(self.hdulist[8].data)
        self.num_center_star_classifications = len(self.center_star_classifications)
        convert_json(self.center_star_classifications, 'center_points')
        convert_json(self.center_star_classifications, 'star_points')
        self.num_center_star_classifications_non_blank = non_blank(self.center_star_classifications, 'center_points_list', 'star_points_list')

    def _process_spiral_classifications(self):
        '''Format the Zooniverse spiral arm classifications as Astropy tables and count the number of unique non-blank classifications'''
        # read in spiral classifications
        self.spiral_classifications = Table(self.hdulist[9].data)
        self.num_spiral_classifications = len(self.spiral_classifications)
        convert_json(self.spiral_classifications, 'spiral_paths')
        self.num_spiral_classifications_non_blank = non_blank(self.spiral_classifications, 'spiral_paths_list')

    def _process_bar_classifications(self):
        '''Format the Zooniverse bar classifications as Astropy tables and count the number of unique non-blank classifications'''
        # read in bar classifications
        self.bar_classifications = Table(self.hdulist[10].data)
        self.num_bar_classifications = len(self.bar_classifications)
        convert_json(self.bar_classifications, 'bar_paths')
        self.num_bar_classifications_non_blank = non_blank(self.bar_classifications, 'bar_paths_list')

    def center_in_pix(self):
        '''Return the center of the IFU in image coordinates'''
        return self.wcs.wcs_world2pix(np.array([[self.metadata['ra'][0], self.metadata['dec'][0]]]), 1)[0]

    def get_hexagon(self, correct_hex=True, edgecolor='magenta'):
        '''Get the IFU hexagon in image as a matplotlib polygon for plotting

        Paramters:
            correct_hex (bool, default=True):
                If True it returns the correct IFU hexagon, if False it returns the hexagon shown
                to the GZ3D volunteers (this was slightly too small due to a bug when producing the
                original images for the project).
            edgecolor (matplotlib color):
                What color to make the hexagon.

        Returns:
            hexagon (matplotlib.patches.RegularPolygon):
                A matplotlib patch object of the IFU hexagon returned in image coordinates.
        '''
        # the spacing should be ~0.5 arcsec not 0, and it should not be rotated by np.sqrt(3) / 2
        if correct_hex:
            # each hex has a total diameter of 2.5 arcsec on the sky (only 2 of it is a fiber)
            diameter = 2.5 / 0.099
            # the radius for mpl is from the center to each vertex, not center to side
            r = LUT[self.ifu_size] * diameter / 2
        else:
            # this was me being wrong about the hexagon params forgetting about the space between fibers
            diameter = 2.0 / 0.099
            # The factor of 1.1 was to try (and fail) to account for the space between fibers :(
            r = 1.1 * LUT[self.ifu_size] * diameter / 2
        c = self.center_in_pix()
        return patches.RegularPolygon(c, 6, r, fill=False, orientation=np.deg2rad(30), edgecolor=edgecolor, linewidth=0.8)

    def _get_ellipse_list(self, table):
        '''Convert table of x, y, var_x, var_y, var_x_y into matplotlib ellipse objects (one for each row)'''
        ellip_list = []
        for idx in range(len(table)):
            pos = np.array([table['x'][idx], table['y'][idx]])
            cov = np.array([[table['var_x'][idx], table['var_x_y'][idx]], [table['var_x_y'][idx], table['var_y'][idx]]])
            ellip_list.append(cov_to_ellipse(cov, pos, nstd=2, edgecolor='k', facecolor='none', lw=1))
        return ellip_list

    def get_center_ellipse_list(self):
        '''Return matplotlib ellipse objects for identified galaxy center(s)'''
        return self._get_ellipse_list(self.center_clusters)

    def get_star_ellipse_list(self):
        '''Return matplotlib ellipse objects for identified forground star(s)'''
        return self._get_ellipse_list(self.star_clusters)

    def _get_spaxel_grid_xy(self, include_edges=False, grid_size=None):
        '''Find the spaxel grid (in image coordinates) for the images in the fits file'''
        if grid_size is None:
            grid_size = self.cube.data['FLUX'].data.shape[1:]
        one_grid = 0.5 / 0.099
        c = self.center_in_pix()
        grid_y = np.arange(grid_size[0] + include_edges) * one_grid
        grid_x = np.arange(grid_size[1] + include_edges) * one_grid
        grid_y = grid_y - np.median(grid_y) + c[0]
        grid_x = grid_x - np.median(grid_x) + c[1]
        return grid_x, grid_y

    def get_spaxel_grid(self, grid_size=None):
        '''Return the data needed to plot the spaxel grid over the GZ3D image'''
        grid_x, grid_y = self._get_spaxel_grid_xy(include_edges=True, grid_size=grid_size)
        v_line_x = np.vstack([grid_x, grid_x])
        v_line_y = np.array([[grid_y[0]], [grid_y[-1]]])
        h_line_x = np.array([[grid_x[0]], [grid_x[-1]]])
        h_line_y = np.vstack([grid_y, grid_y])
        return [(v_line_x, v_line_y), (h_line_x, h_line_y)]

    def _get_spaxel_mask(self, mask, grid_size=None):
        '''Resample GZ3D masks onto the spaxel grid using a Bivariate spline resampling'''
        # assumes a 0.5 arcsec grid centered on the ifu's ra and dec
        # use a Bivariate spline approximation to resample mask to the spaxel grid
        xx = np.arange(mask.shape[1])
        yy = np.arange(mask.shape[0])
        s = RectBivariateSpline(xx, yy, mask)
        grid_x, grid_y = self._get_spaxel_grid_xy(grid_size=grid_size)
        # flip the output mask so the origin is the lower left of the image
        s_mask = np.flipud(s(grid_x, grid_y))
        # zero out small values
        s_mask[s_mask < 0.5] = 0
        return s_mask

    def _process_all_spaxel_masks(self, grid_size=None):
        '''Process all GZ3D masks onto the MaNGA spaxel grid and give them useful names'''
        self.center_mask_spaxel = self._get_spaxel_mask(self.center_mask, grid_size=grid_size)
        self.star_mask_spaxel = self._get_spaxel_mask(self.star_mask, grid_size=grid_size)
        self.spiral_mask_spaxel = self._get_spaxel_mask(self.spiral_mask, grid_size=grid_size)
        self.bar_mask_spaxel = self._get_spaxel_mask(self.bar_mask, grid_size=grid_size)
        self.other_mask_spaxel = (self.spiral_mask_spaxel == 0) & (self.bar_mask_spaxel == 0) & (self.center_mask_spaxel == 0)

    def _stack_spectra(self, mask_name, inv=False):
        '''Stack multiple spectra withing a spaxel mask following Westfall et al. 2019 for covariance correction factors'''
        mask = getattr(self, mask_name)
        if inv:
            mask = mask.max() - mask
        mdx = np.where(mask > 0)
        if len(mdx[0] > 0):
            weights = mask[mdx]
            spaxel_index = np.array(mdx).T

            spectra = [s.flux for s in self.cube[mdx]]

            # only keep spectra inside the IFU
            in_ifu = np.array([not any(2**0 & s.mask) for s in spectra])
            if in_ifu.sum() == 0:
                return None

            spectra = [spectra[i] for i in in_ifu.nonzero()[0]]
            spaxel_index = spaxel_index[in_ifu]
            weights = weights[in_ifu]
            weights_total = weights.sum()

            if len(spectra) == 1:
                return spectra[0]

            flux = np.array([sp.value for sp in spectra])
            # the weighted mean
            mean = (flux * weights[:, None]).sum(axis=0) / weights_total

            # we need to handle covariance between spaxels when calculating
            # uncertainties. We follow Westfall et al. 2019's method based in
            # distance between spaxels
            d = distance_matrix(spaxel_index, spaxel_index) / 1.92
            roh = np.exp(-0.5 * d**2)

            # only work with value where roh > 0.003
            xx, yy = np.where(roh > 0.003)
            sigma = np.array([sp.error.value for sp in spectra])
            running_sum = np.zeros_like(sigma[0])
            for idx, jdx in zip(xx, yy):
                running_sum += roh[idx, jdx] * weights[idx] * weights[jdx] * sigma[idx] * sigma[jdx]
            ivar = (weights_total**2) / running_sum

            return Spectrum(
                mean,
                unit=spectra[0].unit,
                wavelength=spectra[0].wavelength,
                wavelength_unit=spectra[0].wavelength.unit,
                pixmask_flag=spectra[0].pixmask_flag,
                ivar=ivar
            )

        return None

    def get_mean_spectra(self, inv=False):
        '''Calculate the mean spectra inside each of the spaxel masks accounting
        for covariance following Westfall et al. 2019's method based in distance
        between spaxels.

        Parameters:
            inv (bool, default=False):
                If true this function will also calculate the mean spectra
                for each inverted mask.  Useful if you want to make difference
                spectra (e.g. <spiral> - <not spiral>).

        Attributes:
            mean_bar (marvin.tools.quantities.spectrum):
                average spectra inside the bar mask
            mean_spiral (marvin.tools.quantities.spectrum):
                average spectra inside the spiral mask
            mean_center (marvin.tools.quantities.spectrum):
                average spectra inside the center mask
            mean_not_bar (marvin.tools.quantities.spectrum):
                average spectra outside the bar mask
            mean_not_spiral (marvin.tools.quantities.spectrum):
                average spectra outside the spiral mask
            mean_not_center (marvin.tools.quantities.spectrum):
                average spectra outside the center mask
        '''
        if self.mean_bar is None:
            self.mean_bar = self._stack_spectra('bar_mask_spaxel')
        if self.mean_spiral is None:
            self.mean_spiral = self._stack_spectra('spiral_mask_spaxel')
        if self.mean_center is None:
            self.mean_center = self._stack_spectra('center_mask_spaxel')
        if inv:
            if self.mean_not_bar is None:
                self.mean_not_bar = self._stack_spectra('bar_mask_spaxel', inv=True)
            if self.mean_not_spiral is None:
                self.mean_not_spiral = self._stack_spectra('spiral_mask_spaxel', inv=True)
            if self.mean_not_center is None:
                self.mean_not_center = self._stack_spectra('center_mask_spaxel', inv=True)

    def _get_bpt(self, snr_min=3, oi_sf=False):
        '''Use the `bpt` module to grab the information needed to make BPT plots color coded by the GZ3D masks'''
        # Gets the necessary emission line maps
        oiii = bpt.get_masked(self.maps, 'oiii_5008', snr=bpt.get_snr(snr_min, 'oiii'))
        nii = bpt.get_masked(self.maps, 'nii_6585', snr=bpt.get_snr(snr_min, 'nii'))
        ha = bpt.get_masked(self.maps, 'ha_6564', snr=bpt.get_snr(snr_min, 'ha'))
        hb = bpt.get_masked(self.maps, 'hb_4862', snr=bpt.get_snr(snr_min, 'hb'))
        sii = bpt.get_masked(self.maps, 'sii_6718', snr=bpt.get_snr(snr_min, 'sii'))
        oi = bpt.get_masked(self.maps, 'oi_6302', snr=bpt.get_snr(snr_min, 'oi'))
        self.log_oiii_hb = np.ma.log10(oiii / hb)
        self.log_nii_ha = np.ma.log10(nii / ha)
        self.log_sii_ha = np.ma.log10(sii / ha)
        self.log_oi_ha = np.ma.log10(oi / ha)
        sf_mask_nii = ((self.log_oiii_hb < bpt.kewley_sf_nii(self.log_nii_ha)) & (self.log_nii_ha < 0.05)).filled(False)
        sf_mask_sii = ((self.log_oiii_hb < bpt.kewley_sf_sii(self.log_sii_ha)) & (self.log_sii_ha < 0.32)).filled(False)
        sf_mask_oi = ((self.log_oiii_hb < bpt.kewley_sf_oi(self.log_oi_ha)) & (self.log_oi_ha < -0.59)).filled(False)
        if oi_sf:
            self.sf_mask = sf_mask_nii & sf_mask_sii & sf_mask_oi
        else:
            self.sf_mask = sf_mask_nii & sf_mask_sii

    def get_distance(self):
        '''Find the radial distance between each spaxel and the center of the galaxy'''
        if self.dis is None:
            cdx = np.unravel_index(self.center_mask_spaxel.argmax(), self.center_mask_spaxel.shape)
            self.dis = np.zeros_like(self.center_mask_spaxel)
            for yy in range(self.dis.shape[0]):
                for xx in range(self.dis.shape[1]):
                    self.dis[yy, xx] = np.linalg.norm([yy - cdx[0], xx - cdx[1]])

    def _set_up_axes(self, ax, color_grid=None):
        '''Helper function to set RA and DEC ticks on plots'''
        ra = ax.coords['ra']
        dec = ax.coords['dec']
        # add axis labels
        ra.set_axislabel('RA')
        dec.set_axislabel('Dec')
        ra.set_major_formatter('d.ddd')
        ra.ticklabels.set_rotation(90)
        ra.ticklabels.set_rotation_mode('anchor')
        ra.ticklabels.set_pad(15)
        dec.set_major_formatter('d.ddd')
        ra.display_minor_ticks(True)
        dec.display_minor_ticks(True)
        # add a coordinate grid to the image
        if color_grid is not None:
            ax.coords.grid(color=color_grid, alpha=0.5, linestyle='solid', lw=1.5)

    def plot_image(self, ax=None, color_grid=None, correct_hex=True, hex_color='C7'):
        '''Plot original GZ3D image that was shown to volunteers.

        Keywords:
            ax (matplotlib.axes.Axes):
                Matplotlib axis object. This axis must have a WCS projection set e.g.
                `ax = fig.add_subplot(111, projection=data.wcs)`. If not provided a new
                figure and axis will be created with the correct projection.
            color_grid (string):
                A matplotlib color to use for the RA-DEC grid lines. Default `None`.
            correct_hex (bool):
                If set to true the correct MaNGA hexagon will be plotted on top of the
                galaxy cutout (the hexagon in the image shown to the volunteers was slightly
                too small due to a bug when producing the original images for the project).
            hex_color (string):
                A matplotlib color to use for the correct MaNGA hexagon if `correct_hex` is
                True.  Default is `'C7'`.

        Returns:
            ax (matplotlib.axes.Axes):
                The matplotlib axis object for the resulting plot.
        '''
        if (ax is None):
            fig = plt.figure()
            ax = fig.add_subplot(111, projection=self.wcs)
        try:
            self._set_up_axes(ax, color_grid=color_grid)
        except AttributeError:
            raise TypeError('ax must have a WCS project, e.g. `ax = fig.add_subplot(111, projection=data.wcs)`')
        if correct_hex:
            ax.add_patch(self.get_hexagon(correct_hex=True, edgecolor=hex_color))
        ax.imshow(self.image)
        return ax

    def plot_masks(
        self,
        colors=['C1', 'C0', 'C4', 'C2'],
        color_grid=None,
        hex=True,
        hex_color='C7',
        show_image=False,
        subplot_spec=None,
        spaxel_masks=False
    ):
        '''Plot GZ3D masks

        Keywords:
            colors (list):
                A list of matplotlib colors to use for each of the masks. The order of the list is:
                [Bar, Spiral, Forground Stars, Galaxy Center(s)]. Default value is
                `['C1', 'C0', 'C4', 'C2']`.
            color_grid (string):
                A matplotlib color to use for the RA-DEC grid lines. Default `None`.
            hex (bool):
                If `True` plot the MaNGA hexagon. Default value is `True`.
            hex_color (string):
                A matplotlib color to use for the correct MaNGA hexagon if `correct_hex` is
                True.  Default is `'C7'`.
            show_image (bool):
                If `True` plot the original galaxy image behind the masks. Default is `False`.
            subplot_spec (matplotlib.gridspec.SubplotSpec):
                A gridspec subplot specification for this plot. If `None` is provided a new
                figure will be created.
            spaxel_masks (bool):
                If `True` use the masks projected on to the MaNGA spaxel grid, other wise
                plot them on the pixel grid of the GZ3D image shown to the volunteers. Default
                value is `False`.
        '''
        if subplot_spec is None:
            fig = plt.figure()
            # image axis
            gs = gridspec.GridSpec(1, 2, width_ratios=[0.9, 0.1], wspace=0.01)
        else:
            gs = gridspec.GridSpecFromSubplotSpec(1, 2, width_ratios=[0.9, 0.1], wspace=0.01, subplot_spec=subplot_spec)

        # color bar axis
        gs_color_bars = gridspec.GridSpecFromSubplotSpec(1, 4, wspace=0, subplot_spec=gs[1])

        # alpha overlay all masks with correct colors
        if spaxel_masks:
            mask_list = [self.bar_mask_spaxel, self.spiral_mask_spaxel, self.star_mask_spaxel, self.center_mask_spaxel]
            ax1 = plt.subplot(gs[0], projection=self.maps.wcs)
        else:
            mask_list = [self.bar_mask, self.spiral_mask, self.star_mask, self.center_mask]
            ax1 = plt.subplot(gs[0], projection=self.wcs)

        if (show_image) and (not spaxel_masks):
            all_masks = alpha_maps(mask_list, colors, background_image=self.image / 255)
        else:
            all_masks = alpha_maps(mask_list, colors)

        self._set_up_axes(ax1, color_grid=color_grid)
        ax1.imshow(all_masks)

        # overlay IFU hexagon
        if hex:
            ax1.add_patch(self.get_hexagon(correct_hex=True, edgecolor=hex_color))

        # plot center and star ellipses to better define ellipse shape
        if (not spaxel_masks):
            center_ellipse = self.get_center_ellipse_list()
            for e, count in zip(center_ellipse, self.center_clusters['count']):
                e.set_edgecolor(make_alpha_color(count, colors[3]))
                ax1.add_artist(e)
            star_ellipse = self.get_star_ellipse_list()
            for e, count in zip(star_ellipse, self.star_clusters['count']):
                e.set_edgecolor(make_alpha_color(count, colors[2]))
                ax1.add_artist(e)

        # make the legend
        bar_patch = mpl.patches.Patch(color=colors[0], label='Bar')
        spiral_patch = mpl.patches.Patch(color=colors[1], label='Spiral')
        star_patch = mpl.patches.Patch(color=colors[2], label='Star')
        center_patch = mpl.patches.Patch(color=colors[3], label='Center')
        plt.legend(
            handles=[bar_patch, spiral_patch, star_patch, center_patch],
            ncol=2,
            loc='lower center',
            mode='expand'
        )

        # make the colorbars
        ax_bar, cb_bar = plot_alpha_bar(colors[0], gs_color_bars[0])
        ax_spiral, cb_spiral = plot_alpha_bar(colors[1], gs_color_bars[1])
        ax_star, cb_star = plot_alpha_bar(colors[2], gs_color_bars[2])
        ax_center, cb_center = plot_alpha_bar(colors[3], gs_color_bars[3])
        ax_center.tick_params(axis=u'both', which=u'both', length=0)
        tick_labels = np.arange(0, 16)
        tick_locs = tick_labels - 0.5
        cb_center.set_ticks(tick_locs)
        cb_center.set_ticklabels(tick_labels)
        cb_center.set_label('Count')

        return gs

    def _plot_bpt_boundary(self, ax, bpt_kind):
        '''Plot the BPT boundary lines'''
        if bpt_kind == 'log_nii_ha':
            xx_sf_nii = np.linspace(-1.281, 0.045, int(1e4))
            xx_comp_nii = np.linspace(-2, 0.4, int(1e4))
            ax.plot(xx_sf_nii, bpt.kewley_sf_nii(xx_sf_nii), 'k--', zorder=90)
            ax.plot(xx_comp_nii, bpt.kewley_comp_nii(xx_comp_nii), 'k-', zorder=90)
            ax.set_xlim(-2, 0.5)
            ax.set_ylim(-1.5, 1.3)
            ax.set_xlabel(r'log([NII]/H$\alpha$)')
            ax.set_ylabel(r'log([OIII]/H$\beta$)')
        elif bpt_kind == 'log_sii_ha':
            xx_sf_sii = np.linspace(-2, 0.315, int(1e4))
            xx_agn_sii = np.array([-0.308, 1.0])
            ax.plot(xx_sf_sii, bpt.kewley_sf_sii(xx_sf_sii), 'k-', zorder=90)
            ax.plot(xx_agn_sii, bpt.kewley_agn_sii(xx_agn_sii), 'k-', zorder=90)
            ax.set_xlim(-1.5, 0.5)
            ax.set_ylim(-1.5, 1.3)
            ax.set_xlabel(r'log([SII]/H$\alpha$)')
            ax.set_ylabel(r'log([OIII]/H$\beta$)')
        elif bpt_kind == 'log_oi_ha':
            xx_sf_oi = np.linspace(-2.5, -0.7, int(1e4))
            xx_agn_oi = np.array([-1.12, 0.5])
            ax.plot(xx_sf_oi, bpt.kewley_sf_oi(xx_sf_oi), 'k-', zorder=90)
            ax.plot(xx_agn_oi, bpt.kewley_agn_oi(xx_agn_oi), 'k-', zorder=90)
            ax.set_xlim(-2.5, 0.0)
            ax.set_ylim(-1.5, 1.3)
            ax.set_xlabel(r'log([OI]/H$\alpha$)')
            ax.set_ylabel(r'log([OIII]/H$\beta$)')
        else:
            raise AttributeError('bpt_kind must be one of "log_nii_ha", "log_sii_ha", or "log_oi_ha", {0} was given'.format(bpt_kind))

    def plot_bpt(
        self,
        ax=None,
        colors=['C1', 'C0', 'C4', 'C2'],
        bpt_kind='log_nii_ha',
        **kwargs
    ):
        '''Plot a BPT diagram for a galaxy that colors the data points based on the GZ3D masks

        Keywords:
            ax (matplotlib.axes.Axes):
                The matplotlib axis object to use for the plot. If `None` is provided a new
                figure and axis is created for the plot.
            colors (list):
                A list of matplotlib colors to use for each of the masks. The order of the list is:
                [Bar, Spiral, Forground Stars, Galaxy Center(s)]. Default value is
                `['C1', 'C0', 'C4', 'C2']`.
            bpt_kind (string):
                The kind of BPT plot to make. This can be one of three values `'log_nii_ha'` (default),
                `'log_sii_ha'`, or `'log_oi_ha'`.
            kwargs:
                All other keywords are pass forward to matplotlib's scatter plot function.

        Returns:
            ax (matplotlib.axes.Axes):
                The matplotlib axis object for the resulting plot.
        '''
        if bpt_kind not in ["log_nii_ha", "log_sii_ha", "log_oi_ha"]:
            raise AttributeError('bpt_kind must be one of "log_nii_ha", "log_sii_ha", or "log_oi_ha", {0} was given'.format(bpt_kind))
        y = self.log_oiii_hb
        x = getattr(self, bpt_kind)
        mdx = ~(y.mask | x.mask)
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        s = kwargs.pop('s', 8)
        odx = mdx & (self.other_mask_spaxel > 0)
        ax.scatter(x[odx], y[odx], c='#c5c5c5', edgecolor='#c5c5c5', s=s, label='Other', **kwargs)
        plot_alpha_scatter(x, y, self.spiral_mask_spaxel, colors[1], ax, s=s, sf_mask=mdx, snr=None, value=False, label='Spiral', **kwargs)
        plot_alpha_scatter(x, y, self.bar_mask_spaxel, colors[0], ax, s=s, sf_mask=mdx, snr=None, value=False, label='Bar', **kwargs)
        plot_alpha_scatter(x, y, self.star_mask_spaxel, colors[2], ax, s=s, sf_mask=mdx, snr=None, value=False, label='Star', **kwargs)
        plot_alpha_scatter(x, y, self.center_mask_spaxel, colors[3], ax, s=s, sf_mask=mdx, snr=None, value=False, label='Center', **kwargs)
        self._plot_bpt_boundary(ax, bpt_kind)
        return ax

    def polar_plot(
        self,
        x_unit='theta',
        ax=None,
        colors=['C1', 'C0', 'C4', 'C2'],
        key='specindex_dn4000',
        ylabel=r'D_{n}4000',
        snr=3,
        sf_only=False,
        **kwargs
    ):
        '''Make a plot of a MaNGA Map value vs. R or theta with the points color coded by
        what GZ3D mask they belong to.

        x_unit (string):
            What x-value to plot against. Either `'theta'` (default) or `'radius'`.
        ax (matplotlib.axes.Axes):
                The matplotlib axis object to use for the plot. If `None` is provided a new
                figure and axis is created for the plot.
        colors (list):
            A list of matplotlib colors to use for each of the masks. The order of the list is:
            [Bar, Spiral, Forground Stars, Galaxy Center(s)]. Default value is
            `['C1', 'C0', 'C4', 'C2']`.
        key (string):
            Name of the MaNGA Map attribute to plot. The default value is `'specindex_dn4000'`.
        ylabel (string):
            The `ylabel` to use for the plot (units will automatically be added to the label
            based on the map being used).
        snr (float):
            The minimum signal to noise cutoff to use for the plot. The default value is `3`.
        sf_only (bool):
            If `True` only plot spaxes that are star forming. The default value is `False`.
        kwargs:
            All other keywords are pass forward to matplotlib's scatter plot function.

        Returns:
            ax (matplotlib.axes.Axes):
                The matplotlib axis object for the resulting plot.
        '''
        title = []
        s = kwargs.pop('s', 8)
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        if x_unit.lower() == 'theta':
            x = self.maps.spx_ellcoo_elliptical_azimuth.value
            ax.set_xticks([0, 90, 180, 270, 360])
            ax.set_xlabel(r'$\theta$')
        elif x_unit.lower() == 'radius':
            r = self.maps.spx_ellcoo_elliptical_radius.value
            r_50 = self.maps.nsa['elpetro_th50_r']
            x = r / r_50
            ax.set_xlabel(r'R / R$_{50}$')
        else:
            raise AttributeError('x_unit must be either `theta` or `radius`, {0} was given'.format(x_unit))
        line = self.maps[key]
        # other spaxel masks
        odx = (self.other_mask_spaxel > 0) & (line.value > 0)
        if snr is not None:
            title.append('S/N > {0}'.format(snr))
            odx = odx & (line.snr > snr)
        sf_mask = None
        if sf_only:
            # star forming only
            sf_mask = self.sf_mask
            title.append('star forming only')
            odx = odx & sf_mask
        #  plot scatter points
        ax.scatter(x[odx], line.value[odx], c='#c5c5c5', edgecolor='#c5c5c5', s=s, **kwargs, label='Other')
        plot_alpha_scatter(x, line, self.spiral_mask_spaxel, colors[1], ax, s=s, snr=snr, sf_mask=sf_mask, label='Spiral')
        plot_alpha_scatter(x, line, self.bar_mask_spaxel, colors[0], ax, s=s, snr=snr, sf_mask=sf_mask, label='Bar')
        plot_alpha_scatter(x, line, self.star_mask_spaxel, colors[2], ax, s=s, snr=snr, sf_mask=sf_mask, label='Star')
        plot_alpha_scatter(x, line, self.center_mask_spaxel, colors[3], ax, s=s, snr=snr, sf_mask=sf_mask, label='Center')
        if len(title) > 0:
            ax.set_title(','.join(title))
        if line.unit.to_string() != '':
            ax.set_ylabel('$\\mathrm{{{}}}\\,[${}$]$'.format(ylabel, line.unit.to_string('latex')))
        else:
            ax.set_ylabel('$\\mathrm{{{}}}$'.format(ylabel))
        return ax

    def __str__(self):
        '''A useful summary of the data in the GZ3D fits file'''
        return '\n'.join([
            'Subject info:',
            '    subject id: {0}'.format(self.subject_id),
            '    manga id: {0}'.format(self.metadata['MANGAID'][0]),
            '    ra: {0}'.format(self.metadata['ra'][0]),
            '    dec: {0}'.format(self.metadata['dec'][0]),
            '    ifu size: {0}'.format(self.ifu_size),
            'Classification counts:',
            '    {0} center/star, {1} non_blank'.format(self.num_center_star_classifications, self.num_center_star_classifications_non_blank),
            '    {0} spiral, {1} non_blank'.format(self.num_spiral_classifications, self.num_spiral_classifications_non_blank),
            '    {0} bar, {1} non_blank'.format(self.num_bar_classifications, self.num_bar_classifications_non_blank),
            'Cluster counts:',
            '    {0} center(s)'.format(self.num_centers),
            '    {0} star(s)'.format(self.num_stars)
        ])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-08-08
# @Filename: MPL5.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-07-30 11:44:55


from __future__ import absolute_import, division, print_function

from astropy import units as u

from marvin.utils.datamodel.maskbit import get_maskbits

from .base import Bintype, Channel, DAPDataModel, MultiChannelProperty, Property, Template, spaxel


M11_STELIB_ZSOL = Template('M11-STELIB-ZSOL', n=0,
                           description='M11 stellar template, solar metallicity.')
MIUSCAT_THIN = Template('MIUSCAT-THIN', n=1, description='MIUSCAT stellar template.')
MILES_THIN = Template('MILES-THIN', n=2, description='MILES stellar template.')

NONE = Bintype('NONE', binned=False, n=3, description='No binning.')
RADIAL = Bintype('RADIAL', n=7, description='Bin spectra from 0-1 Re and from 1-2 Re in two bins.')
STON = Bintype('STON', n=1, description='Bin to S/N=30; only include S/N>5 spectra; '
                                        'fit V, sigma, h3, h4 for stellar kinematics.')


binid_property = Property('binid', ivar=False, mask=False, channel=None,
                          formats={'string': 'Bin ID'},
                          description='ID number for the bin for which the pixel value was '
                                      'calculated; bins are sorted by S/N.')


MPL4_emline_channels = [
    Channel('oiid_3728', formats={'string': 'OIId 3728',
                                  'latex': r'$\forb{O\,II}\;\lambda\lambda 3726,3728$'}, idx=0,
            description='Single Gaussian fit to the (unresolved) OII 3727 doublet feature'),
    Channel('hb_4862', formats={'string': 'H-beta 4862',
                                'latex': r'H$\beta\;\lambda 4862$'}, idx=1),
    Channel('oiii_4960', formats={'string': 'OIII 4960',
                                  'latex': r'$\forb{O\,III}\;\lambda 4960$'}, idx=2),
    Channel('oiii_5008', formats={'string': 'OIII 5008',
                                  'latex': r'$\forb{O\,III}\;\lambda 5008$'}, idx=3),
    Channel('oi_6302', formats={'string': 'OI 6302',
                                'latex': r'$\forb{O\,I}\;\lambda 6302$'}, idx=4),
    Channel('oi_6365', formats={'string': 'OI 6365',
                                'latex': r'$\forb{O\,I}\;\lambda 6365$'}, idx=5),
    Channel('nii_6549', formats={'string': 'NII 6549',
                                 'latex': r'$\forb{N\,II}\;\lambda 6549$'}, idx=6),
    Channel('ha_6564', formats={'string': 'H-alpha 6564',
                                'latex': r'H$\alpha\;\lambda 6564$'}, idx=7),
    Channel('nii_6585', formats={'string': 'NII 6585',
                                 'latex': r'$\forb{N\,II}\;\lambda 6585$'}, idx=8),
    Channel('sii_6718', formats={'string': 'SII 6718',
                                 'latex': r'$\forb{S\,II}\;\lambda 6718$'}, idx=9),
    Channel('sii_6732', formats={'string': 'SII 6732',
                                 'latex': r'$\forb{S\,II\]\;\lambda 6732$'}, idx=10)
]

MPL4_specindex_channels = [
    Channel('d4000', formats={'string': 'D4000'}, unit=u.dimensionless_unscaled, idx=0),
    Channel('caii0p39', formats={'string': 'CaII 0p39'}, unit=u.Angstrom, idx=1),
    Channel('hdeltaa', formats={'string': 'HDeltaA',
                                'latex': r'H\delta\,A'}, unit=u.Angstrom, idx=2),
    Channel('cn1', formats={'string': 'CN1'}, unit=u.mag, idx=3),
    Channel('cn2', formats={'string': 'CN2'}, unit=u.mag, idx=4),
    Channel('ca4227', formats={'string': 'Ca 4227',
                               'latex': r'Ca\,\lambda 4227'}, unit=u.Angstrom, idx=5),
    Channel('hgammaa', formats={'string': 'HGammaA',
                                'latex': r'H\gamma\,A'}, unit=u.Angstrom, idx=6),
    Channel('fe4668', formats={'string': 'Fe 4668',
                               'latex': r'Fe\,\lambda 4668'}, unit=u.Angstrom, idx=7),
    Channel('hb', formats={'string': 'Hb',
                           'latex': r'H\beta'}, unit=u.Angstrom, idx=8),
    Channel('mgb', formats={'string': 'Mgb'}, unit=u.Angstrom, idx=9),
    Channel('fe5270', formats={'string': 'Fe 5270',
                               'latex': r'Fe\,\lambda 5270'}, unit=u.Angstrom, idx=10),
    Channel('fe5335', formats={'string': 'Fe 5335',
                               'latex': r'Fe\,\lambda 5335'}, unit=u.Angstrom, idx=11),
    Channel('fe5406', formats={'string': 'Fe 5406',
                               'latex': r'Fe\,\lambda 5406'}, unit=u.Angstrom, idx=12),
    Channel('nad', formats={'string': 'NaD'}, unit=u.Angstrom, idx=13),
    Channel('tio1', formats={'string': 'TiO1'}, unit=u.mag, idx=14),
    Channel('tio2', formats={'string': 'TiO2'}, unit=u.mag, idx=15),
    Channel('nai0p82', formats={'string': 'NaI 0p82'}, unit=u.Angstrom, idx=16),
    Channel('caii0p86a', formats={'string': 'CaII 0p86A'}, unit=u.Angstrom, idx=17),
    Channel('caii0p86b', formats={'string': 'CaII 0p86B'}, unit=u.Angstrom, idx=18),
    Channel('caii0p86c', formats={'string': 'CaII 0p86C'}, unit=u.Angstrom, idx=19),
    Channel('mgi0p88', formats={'string': 'MgI 0p88'}, unit=u.Angstrom, idx=20),
    Channel('tio0p89', formats={'string': 'TiO 0p89'}, unit=u.Angstrom, idx=21),
    Channel('feh0p99', formats={'string': 'FeH 0p99'}, unit=u.Angstrom, idx=22)
]


MPL4_maps = [
    MultiChannelProperty('emline_gflux', ivar=True, mask=True, channels=MPL4_emline_channels,
                         unit=u.erg / u.s / (u.cm ** 2) / spaxel, scale=1e-17,
                         formats={'string': 'Emission line Gaussian flux'},
                         description='Fluxes of emission lines based on a single Gaussian fit.'),
    MultiChannelProperty('emline_gvel', ivar=True, mask=True, channels=MPL4_emline_channels,
                         unit=u.km / u.s,
                         formats={'string': 'Emission line Gaussian velocity'},
                         description='Doppler velocity shifts for emission lines relative to '
                                     'the NSA redshift based on a single Gaussian fit.'),
    MultiChannelProperty('emline_gsigma', ivar=True, mask=True, channels=MPL4_emline_channels,
                         unit=u.km / u.s,
                         formats={'string': 'Emission line Gaussian sigma',
                                  'latex': r'Emission line Gaussian $\sigma$'},
                         description='Velocity dispersions of emission lines based on a '
                                     'single Gaussian fit.'),
    MultiChannelProperty('emline_instsigma', ivar=False, mask=False,
                         channels=MPL4_emline_channels,
                         unit=u.km / u.s,
                         formats={'string': 'Emission line instrumental sigma',
                                  'latex': r'Emission line instrumental $\sigma$'},
                         description='Instrumental velocity dispersion at the line centroids '
                                     'for emission lines (based on a single Gaussian fit).'),
    MultiChannelProperty('emline_ew', ivar=True, mask=True, channels=MPL4_emline_channels,
                         unit=u.Angstrom,
                         formats={'string': 'Emission line EW'},
                         description='Equivalent widths for emission lines based on a '
                                     'single Gaussian fit.'),
    MultiChannelProperty('emline_sflux', ivar=True, mask=True, channels=MPL4_emline_channels,
                         unit=u.erg / u.s / (u.cm ** 2) / spaxel, scale=1e-17,
                         formats={'string': 'Emission line summed flux'},
                         description='Fluxes for emission lines based on integrating the '
                                     'flux over a set of passbands.'),
    Property('stellar_vel', ivar=True, mask=True, channel=None,
             unit=u.km / u.s,
             formats={'string': 'Stellar velocity'},
             description='Stellar velocity measurements.'),
    Property('stellar_sigma', ivar=True, mask=True, channel=None,
             unit=u.km / u.s,
             formats={'string': 'Stellar velocity dispersion', 'latex': r'Stellar $\sigma$'},
             description='Stellar velocity dispersion measurements.'),
    MultiChannelProperty('specindex', ivar=True, mask=True,
                         formats={'string': 'Spectral index'},
                         channels=MPL4_specindex_channels,
                         description='Measurements of spectral indices.'),
    binid_property
]


MPL4 = DAPDataModel('1.1.1', aliases=['MPL-4', 'MPL4'], bintypes=[NONE, RADIAL, STON],
                    templates=[M11_STELIB_ZSOL, MIUSCAT_THIN, MILES_THIN],
                    properties=MPL4_maps, bitmasks=get_maskbits('MPL-4'),
                    default_bintype='NONE', default_template='MIUSCAT-THIN',
                    property_table='SpaxelProp', default_binid=binid_property,
                    qual_flag=None)



#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-04-29 01:15:33
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-04-29 01:15:33 by Brian Cherinka
    Last Modified On: 2016-04-29 01:15:33 by Brian

'''
from __future__ import print_function
from __future__ import division
from flask import Blueprint, render_template
from flask_classful import route
from marvin.core.exceptions import MarvinError
from marvin.utils.general import get_random_images
from marvin.web.web_utils import buildImageDict
from marvin.web.controllers import BaseWebView

images = Blueprint("images_page", __name__)


class Random(BaseWebView):
    route_base = '/random/'

    def __init__(self):
        ''' Initialize the route '''
        super(Random, self).__init__('marvin-random')
        self.random = self.base.copy()

    def before_request(self, *args, **kwargs):
        ''' Do these things before a request to any route '''
        super(Random, self).before_request(*args, **kwargs)
        self.reset_dict(self.random)

    @route('/', methods=['GET', 'POST'])
    def index(self):

        # Attempt to retrieve search parameters
        self.random['imnumber'] = 16
        images = []

        # Get random images ; parse out thumbnails ; construct plate-IFUs
        imfiles = None
        try:
            imfiles = get_random_images(num=self.random['imnumber'], release=self._release)
        except (MarvinError, AssertionError) as e:
            self.random['error'] = 'Error: could not get images: {0}'.format(e)
        else:
            urls = [i.url for i in imfiles]
            images = buildImageDict(urls)

        # if image grab failed, make placeholders
        if not imfiles:
            images = buildImageDict(imfiles, test=True, num=self.random['imnumber'])

        # Add images to dict
        self.random['images'] = images

        return render_template('random.html', **self.random)


Random.register(images)






# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-09-03 18:56:38
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-20 12:06:47

from __future__ import absolute_import, division, print_function

from brain.utils.general import compress_data
from brain.utils.general.decorators import public
from flask import Response, jsonify, redirect, stream_with_context, url_for
from flask_classful import route
from marvin import config
from marvin.api.base import arg_validate as av
from marvin.api.base import BaseView
from marvin.core.exceptions import MarvinError
from marvin.tools.query import Query, doQuery
from marvin.utils.datamodel.query.base import bestparams
from marvin.utils.db import get_traceback
from marvin.web.extensions import limiter


def _recombine_args(args):
    ''' Recombine any list keyword args intro strings '''
    newargs = {k: ','.join(v) if isinstance(v, list) else v for k, v in args.items()}
    return newargs


def _run_query(searchfilter, **kwargs):
    ''' Run the query and return the query and results '''

    release = kwargs.pop('release', None)
    kwargs['return_params'] = kwargs.pop('returnparams', None)
    kwargs['default_params'] = kwargs.pop('defaults', None)
    kwargs['return_type'] = kwargs.pop('rettype', None)

    try:
        q, r = doQuery(search_filter=searchfilter, release=release, **kwargs)
    except Exception as e:
        raise MarvinError('Query failed with {0}: {1}'.format(e.__class__.__name__, e))
    else:
        return q, r


def _get_runtime(query):
    ''' Retrive a dictionary of the runtime to pass back in JSON '''
    runtime = {'days': query._run_time.days, 'seconds': query._run_time.seconds, 'microseconds': query._run_time.microseconds}
    return runtime


def _getCubes(searchfilter, **kwargs):
    """Run query locally at Utah and format the output into the full JSON """

    # run the query
    q, r = _run_query(searchfilter, **kwargs)

    # get the subset keywords
    start = kwargs.get('start', None)
    end = kwargs.get('end', None)
    limit = kwargs.get('limit', None)
    returnparams = kwargs.get('returnparams', None)

    # retrieve a subset
    chunk = None
    if start:
        chunk = int(end) - int(start)
        # results = r.getSubset(int(start), limit=chunk)
    chunk = limit if not chunk else limit

    # get results
    results = r.results

    # set up the output
    output = dict(data=results, query=r.showQuery(), chunk=limit,
                  filter=searchfilter, params=q.params, returnparams=returnparams, runtime=_get_runtime(q),
                  queryparams_order=q._query_params_order, count=len(results), totalcount=r.totalcount)
    return output


def gen(query, compression=config.compression, params=None):
    ''' Generator for query results

    Parameters:
        query (obj):
            The SQLalchemy query object
        compresssion (str):
            The type of compression to use, e.g. 'json' or 'msgpack'

    Yields:
        A compressed result row of data to stream to the client

    '''
    for i, row in enumerate(query):
        if i == 0 and params:
            yield compress_data(params, compress_with=compression) + ';\n'
        yield compress_data(row, compress_with=compression) + ';\n'


def _get_column(results, colname, format_type=None):
    ''' Gets a column from a Query

    Parameters:
        results (obj):
            A set of Marvin Results
        colname (str):
            The name of the column to extract
        format_type (str):
            The format of the dictionary

    Returns:
        A list of data for that column
    '''

    column = None
    if format_type == 'list':
        try:
            column = results.getListOf(colname)
        except MarvinError as e:
            raise MarvinError('Cannot get list for column {0}: {1}'.format(colname, e))
    elif format_type in ['listdict', 'dictlist']:
        try:
            if colname == 'None':
                column = results.getDictOf(format_type=format_type)
            else:
                column = results.getDictOf(colname, format_type=format_type)
        except MarvinError as e:
            raise MarvinError('Cannot get dictionary for column {0}: {1}'.format(colname, e))

    return column


def _compressed_response(compression, results):
    ''' Compress the data before sending it back in the Response

    Parameters:
        compression (str):
            The compression type.  Either `json` or `msgpack`.
        results (dict):
            The current response dictionary

    Returns:
        A Flask Response object with the compressed data
    '''

    # pack the data
    mimetype = 'json' if compression == 'json' else 'octet-stream'
    try:
        packed = compress_data(results, compress_with=compression)
    except Exception as e:
        results['error'] = str(e)
        results['traceback'] = get_traceback(asstring=True)

    return Response(stream_with_context(packed), mimetype='application/{0}'.format(mimetype))


class QueryView(BaseView):
    """Class describing API calls related to queries."""
    decorators = [limiter.limit("60/minute")]

    def index(self):
        '''Returns general query info

        .. :quickref: Query; Get general query info

        :query string release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: data message
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/query/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": "this is a query!"
           }

        '''
        self.results['data'] = 'this is a query!'
        self.results['status'] = 1
        return jsonify(self.results)

    @route('/stream/', methods=['GET', 'POST'], endpoint='stream')
    @av.check_args(use_params='query', required='searchfilter')
    def stream(self, args):
        ''' test query generator stream '''

        searchfilter = args.pop('searchfilter', None)
        compression = args.pop('compression', config.compression)
        mimetype = 'json' if compression == 'json' else 'octet-stream'

        release = args.pop('release', None)
        args['return_params'] = args.pop('returnparams', None)
        args['return_type'] = args.pop('rettype', None)
        q = Query(search_filter=searchfilter, release=release, **args)

        output = dict(data=None, chunk=q.limit, query=q.show(),
                      filter=searchfilter, params=q.params, returnparams=q.return_params, runtime=None,
                      queryparams_order=q._query_params_order, count=None, totalcount=None)

        return Response(stream_with_context(gen(q.query, compression=compression, params=q.params)), mimetype='application/{0}'.format(mimetype))

    @route('/cubes/', methods=['GET', 'POST'], endpoint='querycubes')
    @av.check_args(use_params='query', required='searchfilter')
    def cube_query(self, args):
        ''' Performs a remote query

        .. :quickref: Query; Perform a remote query

        :query string release: the release of MaNGA
        :form searchfilter: your string searchfilter expression
        :form params: the list of return parameters
        :form rettype: the string indicating your Marvin Tool conversion object
        :form limit: the limiting number of results to return for large results
        :form sort: a string parameter name to sort on
        :form order: the order of the sort, either ``desc`` or ``asc``
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json list results: the list of results
        :json string query: the raw SQL string of your query
        :json int chunk: the page limit of the results
        :json string filter: the searchfilter used
        :json list returnparams: the list of return parameters
        :json list params: the list of parameters used in the query
        :json list queryparams_order: the list of parameters used in the query
        :json dict runtime: a dictionary of query time (days, minutes, seconds)
        :json int totalcount: the total count of results
        :json int count: the count in the current page of results
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/query/cubes/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5", "searchfilter": "nsa.z<0.1"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "chunk": 100,
              "count": 4,
              "data": [["1-209232",8485,"8485-1901","1901",0.0407447],
                       ["1-209113",8485,"8485-1902","1902",0.0378877],
                       ["1-209191",8485,"8485-12701","12701",0.0234253],
                       ["1-209151",8485,"8485-12702","12702",0.0185246]
                      ],
              "filter": "nsa.z<0.1",
              "params": ["cube.mangaid","cube.plate","cube.plateifu","ifu.name","nsa.z"],
              "query": "SELECT ... FROM ... WHERE ...",
              "queryparams_order": ["mangaid","plate","plateifu","name","z"],
              "returnparams": null,
              "runtime": {"days": 0,"microseconds": 55986,"seconds": 0},
              "totalcount": 4
           }

        '''
        # if return_all is True, perform a redirect to stream
        return_all = args.get('return_all', None)
        if return_all:
            args = _recombine_args(args)
            return redirect(url_for('api.stream', **args))

        searchfilter = args.pop('searchfilter', None)

        try:
            res = _getCubes(searchfilter, **args)
        except MarvinError as e:
            self.results['error'] = str(e)
            self.results['traceback'] = get_traceback(asstring=True)
        else:
            self.results['status'] = 1
            self.update_results(res)

        compression = args.pop('compression', config.compression)
        return _compressed_response(compression, self.results)

    @route('/cubes/columns/', defaults={'colname': None}, methods=['GET', 'POST'], endpoint='getcolumn')
    @route('/cubes/columns/<colname>/', methods=['GET', 'POST'], endpoint='getcolumn')
    @av.check_args(use_params='query', required='searchfilter')
    def query_allcolumn(self, args, colname):
        ''' Retrieves the entire result set for a single column

        .. :quickref: Query; Retrieves the entire result set for a single column

        :query string release: the release of MaNGA
        :form searchfilter: your string searchfilter expression
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json list column: the list of results for the specified column
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/query/cubes/columns/plateifu/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5", "searchfilter": "nsa.z<0.1"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "chunk": 100,
              "count": 4,
              "data": ["8485-1901", "8485-1902", "8485-12701", "7443-12701", "8485-12702"],
           }

        '''
        searchfilter = args.pop('searchfilter', None)
        format_type = args.pop('format_type', 'list')
        colname = args.pop('colname', None)

        try:
            query, results = _run_query(searchfilter, **args)
        except MarvinError as e:
            self.results['error'] = str(e)
            self.results['traceback'] = get_traceback(asstring=True)
        else:
            try:
                column = _get_column(results, colname, format_type=format_type)
            except MarvinError as e:
                self.results['error'] = str(e)
                self.results['traceback'] = get_traceback(asstring=True)
            else:
                self.results['status'] = 1
                self.results['data'] = column
                self.results['runtime'] = _get_runtime(query)

        compression = args.pop('compression', config.compression)
        return _compressed_response(compression, self.results)
        #return Response(json.dumps(self.results), mimetype='application/json')

    @route('/cubes/getsubset/', methods=['GET', 'POST'], endpoint='getsubset')
    @av.check_args(use_params='query', required=['searchfilter', 'start', 'end'])
    def query_getsubset(self, args):
        ''' Remotely grab a subset of results from a query

        .. :quickref: Query; Grab a subset of results from a remote query

        :query string release: the release of MaNGA
        :form searchfilter: your string searchfilter expression
        :form params: the list of return parameters
        :form rettype: the string indicating your Marvin Tool conversion object
        :form start: the starting page index of results you wish to grab
        :form end: the ending page index of the results you wish to grab
        :form limit: the limiting number of results to return for large results
        :form sort: a string parameter name to sort on
        :form order: the order of the sort, either ``desc`` or ``asc``
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json list results: the list of results
        :json string query: the raw SQL string of your query
        :json int chunk: the page limit of the results
        :json string filter: the searchfilter used
        :json list returnparams: the list of return parameters
        :json list params: the list of parameters used in the query
        :json list queryparams_order: the list of parameters used in the query
        :json dict runtime: a dictionary of query time (days, minutes, seconds)
        :json int totalcount: the total count of results
        :json int count: the count in the current page of results
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/query/cubes/getsubset/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5", "searchfilter": "nsa.z<0.1", "start":10, "end":15},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "chunk": 100,
              "count": 4,
              "data": [["1-209232",8485,"8485-1901","1901",0.0407447],
                       ["1-209113",8485,"8485-1902","1902",0.0378877],
                       ["1-209191",8485,"8485-12701","12701",0.0234253],
                       ["1-209151",8485,"8485-12702","12702",0.0185246]
                      ],
              "filter": "nsa.z<0.1",
              "params": ["cube.mangaid","cube.plate","cube.plateifu","ifu.name","nsa.z"],
              "query": "SELECT ... FROM ... WHERE ...",
              "queryparams_order": ["mangaid","plate","plateifu","name","z"],
              "returnparams": null,
              "runtime": {"days": 0,"microseconds": 55986,"seconds": 0},
              "totalcount": 4
           }

        '''
        searchfilter = args.pop('searchfilter', None)

        try:
            res = _getCubes(searchfilter, **args)
        except MarvinError as e:
            self.results['error'] = str(e)
            self.results['traceback'] = get_traceback(asstring=True)
        else:
            self.results['status'] = 1
            self.update_results(res)

        compression = args.pop('compression', config.compression)
        return _compressed_response(compression, self.results)
        # this needs to be json.dumps until sas-vm at Utah updates to 2.7.11
        #return Response(json.dumps(self.results), mimetype='application/json')

    @public
    @route('/getparamslist/', methods=['GET', 'POST'], endpoint='getparams')
    @av.check_args(use_params='query', required='paramdisplay')
    def getparamslist(self, args):
        ''' Retrieve a list of all available input parameters into the query

        .. :quickref: Query; Get a list of all or "best" queryable parameters

        :query string release: the release of MaNGA
        :form paramdisplay: ``all`` or ``best``, type of parameters to return
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json list params: the list of queryable parameters
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/query/getparamslist/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": ['nsa.z', 'cube.ra', 'cube.dec', ...]
           }

        '''
        paramdisplay = args.pop('paramdisplay', 'all')
        if paramdisplay == 'all':
            try:
                params = Query.get_available_params('all', release=args['release'])
            except MarvinError as e:
                self.results['error'] = str(e)
                self.results['traceback'] = get_traceback(asstring=True)
                self.results['status'] = -1
                return jsonify(self.results)
        elif paramdisplay == 'best':
            params = bestparams
        self.results['data'] = params
        self.results['status'] = 1
        output = jsonify(self.results)
        return output

    @public
    @route('/getallparams/', methods=['GET', 'POST'], endpoint='getallparams')
    def getall(self):
        ''' Retrieve all the query parameters for all releases at once '''
        params = {}
        releases = config._allowed_releases.keys()
        for release in releases:
            params[release] = Query.get_available_params('all', release=release)
        self.results['data'] = params
        self.results['status'] = 1
        output = jsonify(self.results)
        return output

    @route('/cleanup/', methods=['GET', 'POST'], endpoint='cleanupqueries')
    @av.check_args(use_params='query', required='task')
    def cleanup(self, args):
        ''' Clean up idle server-side queries or retrieve the list of them

        Do not use!

        .. :quickref: Query; Send a cleanup command to the server-side database

        :query string release: the release of MaNGA
        :form task: ``clean`` or ``getprocs``, the type of task to run
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: dictionary of returned data
        :json string clean: clean success message
        :json list procs: the list of processes currently running on the db
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/query/cleanup/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": 'clean success'
           }

        '''
        task = args.pop('task', None)
        if task == 'clean':
            q = Query(mode='local')
            q._cleanUpQueries()
            res = {'status': 1, 'data': 'clean success'}
        elif task == 'getprocs':
            q = Query(mode='local')
            procs = q._getIdleProcesses()
            procs = [{k: v for k, v in y.items()} for y in procs]
            res = {'status': 1, 'data': procs}
        else:
            res = {'status': -1, 'data': None, 'error': 'Task is None or not in [clean, getprocs]'}
        self.update_results(res)
        output = jsonify(self.results)
        return output



#!/usr/bin/env python
# encoding: utf-8
#
# @Author: Brian Cherinka, José Sánchez-Gallego, Brett Andrews
# @Date: Oct 25, 2017
# @Filename: base.py
# @License: BSD 3-Clause
# @Copyright: Brian Cherinka, José Sánchez-Gallego, Brett Andrews


from __future__ import absolute_import, division, print_function

from astropy import units as u

from marvin.utils.datamodel.maskbit import get_maskbits

from .base import RSS, DataCube, DRPCubeDataModel, DRPCubeDataModelList, Spectrum


spaxel_unit = u.Unit('spaxel', represents=u.pixel, doc='A spectral pixel', parse_strict='silent')
fiber_unit = u.Unit('fiber', represents=u.pixel, doc='Spectroscopic fibre', parse_strict='silent')


MPL4_datacubes = [
    DataCube('flux', 'FLUX', 'WAVE', extension_ivar='IVAR',
             extension_mask='MASK', unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
             scale=1e-17, formats={'string': 'Flux'},
             description='3D rectified cube')
]

MPL4_spectra = [
    Spectrum('spectral_resolution', 'SPECRES', extension_wave='WAVE', extension_std='SPECRESD',
             unit=u.Angstrom, scale=1, formats={'string': 'Median spectral resolution'},
             description='Median spectral resolution as a function of wavelength '
                         'for the fibers in this IFU'),
]

MPL6_datacubes = [
    DataCube('dispersion', 'DISP', 'WAVE', extension_ivar=None,
             extension_mask='MASK', unit=u.Angstrom,
             scale=1, formats={'string': 'Dispersion'},
             description='Broadened dispersion solution (1sigma LSF)'),
    DataCube('dispersion_prepixel', 'PREDISP', 'WAVE', extension_ivar=None,
             extension_mask='MASK', unit=u.Angstrom,
             scale=1, formats={'string': 'Dispersion pre-pixel'},
             description='Broadened pre-pixel dispersion solution (1sigma LSF)')

]

MPL6_spectra = [
    Spectrum('spectral_resolution_prepixel', 'PRESPECRES', extension_wave='WAVE',
             extension_std='PRESPECRESD', unit=u.Angstrom, scale=1,
             formats={'string': 'Median spectral resolution pre-pixel'},
             description='Median pre-pixel spectral resolution as a function of '
                         'wavelength for the fibers in this IFU'),
]

RSS_extensions = [
    RSS('xpos', 'XPOS', extension_wave='WAVE', unit=u.arcsec, db_table='rssfiber',
        formats={'string': 'Fiber X-positions from the IFU center'},
        description='Array of fiber X-positions relative to the IFU center'),
    RSS('ypos', 'YPOS', extension_wave='WAVE', unit=u.arcsec, db_table='rssfiber',
        formats={'string': 'Fiber Y-positions from the IFU center'},
        description='Array of fiber Y-positions relative to the IFU center'),
]

MPL10_datacubes = [
    DataCube('dispersion', 'LSFPOST', 'WAVE', extension_ivar=None,
             extension_mask='MASK', unit=u.Angstrom, db_column='disp',
             scale=1, formats={'string': 'Dispersion'},
             description='Broadened dispersion solution (1sigma LSF)'),
    DataCube('dispersion_prepixel', 'LSFPRE', 'WAVE', extension_ivar=None,
             extension_mask='MASK', unit=u.Angstrom, db_column='predisp',
             scale=1, formats={'string': 'Dispersion pre-pixel'},
             description='Broadened pre-pixel dispersion solution (1sigma LSF)')
]

MPL4 = DRPCubeDataModel('MPL-4', aliases=['MPL4', 'v1_5_1'],
                        datacubes=MPL4_datacubes,
                        spectra=MPL4_spectra,
                        bitmasks=get_maskbits('MPL-4'),
                        qual_flag='DRP3QUAL')

MPL5 = DRPCubeDataModel('MPL-5', aliases=['MPL5', 'v2_0_1'],
                        datacubes=MPL4_datacubes,
                        spectra=MPL4_spectra,
                        bitmasks=get_maskbits('MPL-5'),
                        qual_flag='DRP3QUAL')

MPL6 = DRPCubeDataModel('MPL-6', aliases=['MPL6', 'v2_3_1'],
                        datacubes=MPL4_datacubes + MPL6_datacubes,
                        spectra=MPL4_spectra + MPL6_spectra,
                        bitmasks=get_maskbits('MPL-6'),
                        qual_flag='DRP3QUAL')

MPL7 = DRPCubeDataModel('MPL-7', aliases=['MPL7', 'v2_4_3', 'DR15'],
                        datacubes=MPL4_datacubes + MPL6_datacubes,
                        spectra=MPL4_spectra + MPL6_spectra,
                        bitmasks=get_maskbits('MPL-7'),
                        qual_flag='DRP3QUAL')

DR15 = DRPCubeDataModel('DR15', aliases=['DR15', 'v2_4_3'],
                        datacubes=MPL4_datacubes + MPL6_datacubes,
                        spectra=MPL4_spectra + MPL6_spectra,
                        bitmasks=get_maskbits('MPL-7'),
                        qual_flag='DRP3QUAL')

MPL8 = DRPCubeDataModel('MPL-8', aliases=['MPL8', 'v2_5_3'],
                        datacubes=MPL4_datacubes + MPL6_datacubes,
                        spectra=MPL4_spectra + MPL6_spectra,
                        bitmasks=get_maskbits('MPL-8'),
                        qual_flag='DRP3QUAL')

DR16 = DRPCubeDataModel('DR16', aliases=['DR15', 'v2_4_3'],
                        datacubes=MPL4_datacubes + MPL6_datacubes,
                        spectra=MPL4_spectra + MPL6_spectra,
                        bitmasks=get_maskbits('MPL-7'),
                        qual_flag='DRP3QUAL')

MPL9 = DRPCubeDataModel('MPL-9', aliases=['MPL9', 'v2_7_1'],
                        datacubes=MPL4_datacubes + MPL6_datacubes,
                        spectra=MPL4_spectra + MPL6_spectra,
                        bitmasks=get_maskbits('MPL-9'),
                        qual_flag='DRP3QUAL')

MPL10 = DRPCubeDataModel('MPL-10', aliases=['MPL10', 'v3_0_1'],
                         datacubes=MPL4_datacubes + MPL10_datacubes,
                         spectra=MPL4_spectra + MPL6_spectra,
                         bitmasks=get_maskbits('MPL-10'),
                         qual_flag='DRP3QUAL')

MPL11 = DRPCubeDataModel('MPL-11', aliases=['MPL11', 'v3_1_1', 'DR17'],
                         datacubes=MPL4_datacubes + MPL10_datacubes,
                         spectra=MPL4_spectra + MPL6_spectra,
                         bitmasks=get_maskbits('MPL-11'),
                         qual_flag='DRP3QUAL')

# The DRP Cube Datamodel
datamodel = DRPCubeDataModelList([MPL4, MPL5, MPL6, MPL7, DR15, MPL8, DR16, MPL9, MPL10, MPL11])

# Define the RSS Datamodel. Start by copying the Cube datamodel for convenience.
datamodel_rss = datamodel.copy()

for release in datamodel_rss:
    datamodel_rss[release] = datamodel_rss[release].to_rss()
    datamodel_rss[release].rss += RSS_extensions

    flux = datamodel_rss[release].rss.flux
    flux.description = 'Row-stacked spectra from all exposures for the target'
    flux.unit = flux.unit * spaxel_unit / fiber_unit

    locals()



#!/usr/bin/env python
# encoding: utf-8
#
# modelcube.py
#
# Licensed under a 3-clause BSD license.
#
# Revision history:
#     25 Sep 2016 J. Sánchez-Gallego
#       Initial version


from __future__ import division, print_function

import json
import os

from flask import Response, jsonify
from flask_classful import route

from marvin import config
from marvin.api.base import BaseView
from marvin.api.base import arg_validate as av
from marvin.core.exceptions import MarvinError
from marvin.tools.modelcube import ModelCube
from marvin.utils.general import mangaid2plateifu, parseIdentifier


try:
    from sdss_access.path import Path
except ImportError:
    Path = None


def _get_model_cube(name, use_file=False, release=None, **kwargs):
    """Retrieves a Marvin ModelCube object."""

    model_cube = None
    results = {}

    drpver, dapver = config.lookUpVersions(release)

    # parse name into either mangaid or plateifu
    try:
        idtype = parseIdentifier(name)
    except Exception as err:
        results['error'] = 'Failed to parse input name {0}: {1}'.format(name, str(err))
        return model_cube, results

    filename = None
    plateifu = None
    mangaid = None

    bintype = kwargs.pop('bintype')
    template = kwargs.pop('template')

    try:
        if use_file:

            if idtype == 'mangaid':
                plate, ifu = mangaid2plateifu(name, drpver=drpver)
            elif idtype == 'plateifu':
                plate, ifu = name.split('-')

            if Path is not None:

                daptype = '{0}-{1}'.format(bintype, template)

                filename = Path(release=release).full('mangadap', ifu=ifu,
                                                      drpver=drpver,
                                                      dapver=dapver,
                                                      plate=plate, mode='LOGCUBE',
                                                      daptype=daptype)
                assert os.path.exists(filename), 'file not found.'
            else:
                raise MarvinError('cannot create path for MaNGA cube.')

        else:

            if idtype == 'plateifu':
                plateifu = name
            elif idtype == 'mangaid':
                mangaid = name
            else:
                raise MarvinError('invalid plateifu or mangaid: {0}'.format(idtype))

        model_cube = ModelCube(filename=filename, mangaid=mangaid, plateifu=plateifu,
                               release=release, template=template, bintype=bintype, **kwargs)

        results['status'] = 1

    except Exception as err:

        results['error'] = 'Failed to retrieve ModelCube {0}: {1}'.format(name, str(err))

    return model_cube, results


class ModelCubeView(BaseView):
    """Class describing API calls related to ModelCubes."""

    route_base = '/modelcubes/'

    @route('/<name>/', defaults={'bintype': None, 'template': None},
           methods=['GET', 'POST'], endpoint='getModelCube')
    @route('/<name>/<bintype>/', defaults={'template': None},
           methods=['GET', 'POST'], endpoint='getModelCube')
    @route('/<name>/<bintype>/<template>/', methods=['GET', 'POST'], endpoint='getModelCube')
    @av.check_args()
    def get(self, args, name, bintype, template):
        """Retrieves a ModelCube.

        .. :quickref: ModelCube; Get a modelcube given a name, bintype, and template

        :param name: The name of the modelcube as plate-ifu or mangaid
        :param bintype: The bintype associated with this modelcube.
                        If not defined, the default is used
        :param template: The template associated with this modelcube.
                         If not defined, the default is used
        :form release: the release of MaNGA data
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string header: the modelcube header as a string
        :json list shape: the modelcube shape [x, y]
        :json list wavelength: the modelcube wavelength array
        :json list redcorr: the modelcube redcorr array
        :json string wcs_header: the modelcube wcs_header as a string
        :json string bintype: the bintype of the modelcube
        :json string template: the template library of the modelcube
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/modelcubes/8485-1901/SPX/GAU-MILESHC/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"plateifu": 8485-1901,
                       "mangaid": '1-209232',
                       "header": "XTENSION= 'IMAGE', NAXIS=3, .... END",
                       "wcs_header": "WCSAXES = 3 / Number of coordindate axes .... END",
                       "shape": [34, 34],
                       "wavelength": [3621.6, 3622.43, 3623.26, ...],
                       "redcorr": [1.06588, 1.065866, 1.06585, ...],
                       "bintype": "SPX",
                       "template": "GAU-MILESHC"
                      }
           }

        """

        # Pop any args we don't want going into ModelCube
        args = self._pop_args(args, arglist='name')

        model_cube, results = _get_model_cube(name, **args)

        self.update_results(results)

        if model_cube is not None:
            self.results['data'] = {
                'plateifu': model_cube.plateifu,
                'mangaid': model_cube.mangaid,
                'header': model_cube.header.tostring(),
                'shape': model_cube._shape,
                'wavelength': model_cube._wavelength.tolist(),
                'redcorr': model_cube._redcorr.tolist(),
                'wcs_header': model_cube.wcs.to_header_string(),
                'bintype': model_cube.bintype.name,
                'template': model_cube.template.name}

        return jsonify(self.results)

    @route('/<name>/extensions/<modelcube_extension>/',
           defaults={'bintype': None, 'template': None},
           methods=['GET', 'POST'], endpoint='getModelCubeExtension')
    @route('/<name>/<bintype>/extensions/<modelcube_extension>/',
           defaults={'template': None},
           methods=['GET', 'POST'], endpoint='getModelCubeExtension')
    @route('/<name>/<bintype>/<template>/extensions/<modelcube_extension>/',
           methods=['GET', 'POST'], endpoint='getModelCubeExtension')
    @av.check_args()
    def getModelCubeExtension(self, args, name, bintype, template, modelcube_extension):
        """Returns the extension for a modelcube given a plateifu/mangaid.

        .. :quickref: ModelCube; Gets the extension given a plate-ifu or mangaid

        :param name: The name of the cube as plate-ifu or mangaid
        :param bintype: The bintype associated with this modelcube.
        :param template: The template associated with this modelcube.
        :param modelcube_extension: The name of the cube extension.  Either flux, ivar, or mask.
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string modelcube_extension: the data for the specified extension
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/modelcubes/8485-1901/extensions/flux/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"extension_data": [[0,0,..0], [], ... [0, 0, 0,... 0]]
              }
           }
        """

        # Pass the args in and get the cube
        args = self._pop_args(args, arglist=['name', 'modelcube_extension'])
        modelcube, res = _get_model_cube(name, use_file=True, **args)
        self.update_results(res)

        if modelcube:

            extension_data = modelcube.data[modelcube_extension.upper()].data

            if extension_data is None:
                self.results['data'] = {'extension_data': None}
            else:
                self.results['data'] = {'extension_data': extension_data.tolist()}

        return Response(json.dumps(self.results), mimetype='application/json')

    @route('/<name>/binids/<modelcube_extension>/',
           defaults={'bintype': None, 'template': None},
           methods=['GET', 'POST'], endpoint='getModelCubeBinid')
    @route('/<name>/<bintype>/binids/<modelcube_extension>/',
           defaults={'template': None},
           methods=['GET', 'POST'], endpoint='getModelCubeBinid')
    @route('/<name>/<bintype>/<template>/binids/<modelcube_extension>/',
           methods=['GET', 'POST'], endpoint='getModelCubeBinid')
    @av.check_args()
    def getModelCubeBinid(self, args, name, bintype, template, modelcube_extension):
        """Returns the binid array for a modelcube given a plateifu/mangaid.

        .. :quickref: ModelCube; Gets the binid array given a plate-ifu or mangaid

        :param name: The name of the modelcube as plate-ifu or mangaid
        :param bintype: The bintype associated with this modelcube.
        :param template: The template associated with this modelcube.
        :param modelcube_extension: The name of the cube extension.
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string binid: the binid data
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/modelcubes/8485-1901/binids/flux/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"binid": [[0,0,..0], [], ... [0, 0, 0,... 0]]
              }
           }
        """

        # Pass the args in and get the cube
        args = self._pop_args(args, arglist=['name', 'modelcube_extension'])
        modelcube, res = _get_model_cube(name, use_file=False, **args)
        self.update_results(res)

        if modelcube:
            try:
                model = modelcube.datamodel.from_fits_extension(modelcube_extension)
                binid_data = modelcube.get_binid(model)
                self.results['data'] = {'binid': binid_data.value.tolist()}
            except Exception as ee:
                self.results['error'] = str(ee)

        return Response(json.dumps(self.results), mimetype='application/json')

    @route('/<name>/<bintype>/<template>/quantities/<x>/<y>/',
           methods=['GET', 'POST'], endpoint='getModelCubeQuantitiesSpaxel')
    @av.check_args()
    def getModelCubeQuantitiesSpaxel(self, args, name, bintype, template, x, y):
        """Returns a dictionary with all the quantities.

        .. :quickref: ModelCube; Returns a dictionary with all the quantities

        :param name: The name of the cube as plate-ifu or mangaid
        :param bintype: The bintype associated with this modelcube.
        :param template: The template associated with this modelcube.
        :param x: The x coordinate of the spaxel (origin is ``lower``)
        :param y: The y coordinate of the spaxel (origin is ``lower``)
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/modelcubes/8485-1901/SPX/GAU_MILESHC/quantities/10/12/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"binned_flux": {"value": [0,0,..0], "ivar": ...},
                       "emline_fit": ...}
              }
           }
        """

        # Pass the args in and get the cube
        args = self._pop_args(args, arglist=['name', 'x', 'y'])
        modelcube, res = _get_model_cube(name, **args)
        self.update_results(res)

        if modelcube:

            self.results['data'] = {}

            spaxel_quantities = modelcube._get_spaxel_quantities(x, y)

            for quant in spaxel_quantities:

                spectrum = spaxel_quantities[quant]

                if spectrum is None:
                    self.data[quant] = {'value': None}
                    continue

                value = spectrum.value.tolist()
                ivar = spectrum.ivar.tolist() if spectrum.ivar is not None else None
                mask = spectrum.mask.tolist() if spectrum.mask is not None else None

                self.results['data'][quant] = {'value': value,
                                               'ivar': ivar,
                                               'mask': mask}

            self.results['data']['wavelength'] = modelcube._wavelength.tolist()

        return Response(json.dumps(self.results), mimetype='application/json')



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-10-11 17:51:43
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-29 17:23:15

from __future__ import print_function, division, absolute_import

import numpy as np
import astropy
import astropy.units as u
import marvin.tools
from marvin.tools.quantities.spectrum import Spectrum
from marvin.utils.general.general import get_drpall_table
from marvin.utils.plot.scatter import plot as scatplot
from marvin import log

from .base import VACMixIn, VACTarget

def choose_best_spectrum(par1, par2, conf_thresh=0.1):
    '''choose optimal HI spectrum based on the following criteria:
    (1) If both detected and unconfused, choose highest SNR
    (2) If both detected and both confused, choose lower confusion prob.
    (3) If both detected and one confused, choose non-confused
    (4) If one non-confused detection and one non-detection, go with detection
    (5) If one confused detetion and one non-detection, go with non-detection
    (6) If niether detected, choose lowest rms

    par1 and par2 are dictionaries with the following parameters:
    program - gbt or alfalfa
    snr - integrated SNR
    rms - rms noise level
    conf_prob - confusion probability

    conf_thresh = maximum confusion probability below which we classify
    the object as essentially unconfused. Default to 0.1 following
    (Stark+21)

    '''

    programs = [par1['program'],par2['program']]
    sel_high_snr = np.argmax([par1['snr'],par2['snr']])
    sel_low_rms = np.argmin([par1['rms'],par2['rms']])
    sel_low_conf = np.argmin([par1['conf_prob'],par2['conf_prob']])
    

     #both detected
    if (par1['snr'] > 0) & (par2['snr'] > 0):
        if (par1['conf_prob'] <= conf_thresh) & (par2['conf_prob'] <= conf_thresh):
            pick = sel_high_snr
        elif (par1['conf_prob'] <= conf_thresh) & (par2['conf_prob'] > conf_thresh):
            pick = 0
        elif (par1['conf_prob'] > conf_thresh) & (par2['conf_prob'] <= conf_thresh):
             pick = 1
        elif (par1['conf_prob'] > conf_thresh) & (par2['conf_prob'] > conf_thresh):
            pick = sel_low_conf

    #both nondetected
    elif (par1['snr'] <= 0) & (par2['snr'] <= 0):
        pick = sel_low_rms

    #one detected
    elif (par1['snr'] > 0) & (par2['snr'] <= 0):
        if par1['conf_prob'] < conf_thresh:
            pick=0
        else:
            pick=1
    elif (par1['snr'] <= 0) & (par2['snr'] > 0):
        if par2['conf_prob'] < conf_thresh:
            pick=1
        else:
            pick=0

    return programs[pick]
            
class HIVAC(VACMixIn):
    """Provides access to the MaNGA-HI VAC.

    VAC name: HI

    URL: https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=hi-manga-data-release-1

    Description: Returns HI summary data and spectra

    Authors: David Stark and Karen Masters

    """

    # Required parameters
    name = 'HI'
    description = 'Returns HI summary data and spectra'
    version = {'MPL-7': 'v1_0_1', 'DR15': 'v1_0_1', 'DR16': 'v1_0_2', 'DR17': 'v2_0_1', 'MPL-11': 'v2_0_1'}
    display_name = 'HI'
    url = 'https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=hi-manga-data-release-1'

    # optional Marvin Tools to attach your vac to
    include = (marvin.tools.cube.Cube, marvin.tools.maps.Maps, marvin.tools.modelcube.ModelCube)

    # optional methods to attach to your main VAC tool in ~marvin.tools.vacs.VACs
    add_methods = ['plot_mass_fraction']

    # Required method
    def set_summary_file(self, release):
        ''' Sets the path to the HI summary file '''

        # define the variables to build a unique path to your VAC file
        self.path_params = {'ver': self.version[release], 'type': 'all', 'program': 'GBT16A_095'}

        # get_path returns False if the files do not exist locally
        self.summary_file = self.get_path("mangahisum", path_params=self.path_params)

    def set_program(self,plateifu):
        # download the vac from the SAS if it does not already exist locally
        if not self.file_exists(self.summary_file):
            self.summary_file = self.download_vac('mangahisum', path_params=self.path_params)

        # Find all entries in summary file with this plate-ifu. 
        # Need the full summary file data. 
        # Find best entry between GBT/ALFALFA based on dept and confusion. 
        # Then update self.path_params['program'] with alfalfa or gbt.
        
        summary = HITarget(plateifu, vacfile=self.summary_file)._data
        galinfo = summary[summary['plateifu'] == plateifu]
        if len(galinfo) == 1 and galinfo['session']=='ALFALFA':
            program = 'alfalfa'
        elif len(galinfo) in [0, 1]:
            # if no entry found or session is GBT, default program to gbt
            program = 'gbt'
        else:
            par1 = {'program': 'gbt','snr': 0.,'rms': galinfo[0]['rms'], 'conf_prob': galinfo[0]['conf_prob']}
            par2 = {'program': 'gbt','snr': 0.,'rms': galinfo[1]['rms'], 'conf_prob': galinfo[1]['conf_prob']}
            if galinfo[0]['session']=='ALFALFA':
                par1['program'] = 'alfalfa'
            if galinfo[1]['session']=='ALFALFA':
                par2['program'] = 'alfalfa'
            if galinfo[0]['fhi'] > 0:
                par1['snr'] = galinfo[0]['fhi']/galinfo[0]['efhi']
            if galinfo[1]['fhi'] > 0:
                par2['snr'] = galinfo[1]['fhi']/galinfo[1]['efhi']

            program = choose_best_spectrum(par1,par2)

        log.info('Using HI data from {0}'.format(program))
            
        # get path to ancillary VAC file for target HI spectra
        self.update_path_params({'program':program})
        
    # Required method
    def get_target(self, parent_object):
        ''' Accesses VAC data for a specific target from a Marvin Tool object '''

        # get any parameters you need from the parent object
        plateifu = parent_object.plateifu
        self.update_path_params({'plateifu': plateifu})

        
        if parent_object.release in ['DR17', 'MPL-11']:
            self.set_program(plateifu)

        specfile = self.get_path('mangahispectra', path_params=self.path_params)
        
        # create container for more complex return data
        hidata = HITarget(plateifu, vacfile=self.summary_file, specfile=specfile)

        # get the spectral data for that row if it exists
        if hidata._indata and not self.file_exists(specfile):
            hidata._specfile = self.download_vac('mangahispectra', path_params=self.path_params)

        return hidata


class HITarget(VACTarget):
    ''' A customized target class to also display HI spectra

    This class handles data from both the HI summary file and the
    individual spectral files.  Row data from the summary file for the given target
    is returned via the `data` property.  Spectral data can be displayed via
    the the `plot_spectrum` method.

    Parameters:
        targetid (str):
            The plateifu or mangaid designation
        vacfile (str):
            The path of the VAC summary file
        specfile (str):
            The path to the HI spectra

    Attributes:
        data:
            The target row data from the main VAC file
        targetid (str):
            The target identifier
    '''

    def __init__(self, targetid, vacfile, specfile=None):
        super(HITarget, self).__init__(targetid, vacfile)
        self._specfile = specfile
        self._specdata = None

    def plot_spectrum(self):
        ''' Plot the HI spectrum '''

        if self._specfile:
            if not self._specdata:
                self._specdata = self._get_data(self._specfile)
                
            vel = self._specdata['VHI'][0]
            flux = self._specdata['FHI'][0]
            spec = Spectrum(flux, unit=u.Jy, wavelength=vel,
                            wavelength_unit=u.km / u.s)
            ax = spec.plot(
                ylabel='HI\ Flux\ Density', xlabel='Velocity', title=self.targetid, ytrim='minmax'
            )
            return ax
        return None

#
# Functions to become available on your VAC in marvin.tools.vacs.VACs


def plot_mass_fraction(vacdata_object):
    ''' Plot the HI mass fraction
    
    Computes and plots the HI mass fraction using
    the NSA elliptical Petrosian stellar mass from the
    MaNGA DRPall file.  Only plots data for subset of
    targets in both the HI VAC and the DRPall file.

    Parameters:
        vacdata_object (object):
            The `~.VACDataClass` instance of the HI VAC 

    Example:
        >>> from marvin.tools.vacs import VACs
        >>> v = VACs()
        >>> hi = v.HI
        >>> hi.plot_mass_fraction()
    '''
    drpall = get_drpall_table()
    drpall.add_index('plateifu')
    data = vacdata_object.data[1].data
    subset = drpall.loc[data['plateifu']]
    log_stmass = np.log10(subset['nsa_elpetro_mass'])
    diff = data['logMHI'] - log_stmass
    fig, axes = scatplot(
        log_stmass,
        diff,
        with_hist=False,
        ylim=[-5, 5],
        xlabel=r'log $M_*$',
        ylabel=r'log $M_{HI}/M_*$',
    )
    return axes[0]



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-17 23:36:37
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-19 15:44:42

from __future__ import print_function, division, absolute_import

from collections import defaultdict
from marvin.utils.datamodel.drp import datamodel
from marvin.contrib.vacs.base import VACMixIn
from .base import VACList, VACDataModel


subvacs = VACMixIn.__subclasses__()
vacdms = []

# create a dictionary of VACs by release
vacdict = defaultdict(list)
for sv in subvacs:
    # skip hidden VACs
    if sv._hidden:
        continue
    # add versions to dictionary
    for k in sv.version.keys():
        vacdict[k].append(sv)

# create VAC datamodels
for release, vacs in vacdict.items():
    vc = VACList(vacs)
    dm = datamodel[release] if release in datamodel else None
    aliases = dm.aliases if dm else None
    vacdm = VACDataModel(release, vacs=vc, aliases=aliases)
    vacdms.append(vacdm)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-09-28 16:21:17
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-04-24 13:26:54

from __future__ import print_function, division, absolute_import
from flask import Blueprint, render_template
from marvin.web.controllers import BaseWebView


users = Blueprint("users_page", __name__)


class User(BaseWebView):
    route_base = '/user/'

    def __init__(self):
        ''' Initialize the route '''
        super(User, self).__init__('marvin-user')
        self.user = self.base.copy()

    def index(self):
        return render_template('preferences.html', **self.user)

    def preferences(self):
        return render_template('preferences.html', **self.user)

User.register(users)




from .base import *

from .MPL4 import MPL4
from .MPL5 import MPL5
from .MPL6 import MPL6
from .MPL7 import MPL7
from .MPL8 import MPL8
from .MPL9 import MPL9
from .MPL10 import MPL10
from .MPL11 import MPL11

# Defines the list of datamodels.
datamodel = DAPDataModelList([MPL4, MPL5, MPL6, MPL7, MPL8, MPL9, MPL10, MPL11])

